from kvirt import common
from kvirt.config import Kbaseconfig, Kconfig
from kvirt.common import get_git_version, compare_git_versions
from kvirt.nameutils import get_random_name
from mcp.server.fastmcp import FastMCP
from typing import Optional
from urllib.request import urlopen


def _parse_vms_list(_list, overrides={}):
    if isinstance(_list, str):
        print(_list)
        return
    field = overrides.get('field')
    if field is not None:
        vmstable = ["Name", field.capitalize()]
    else:
        vmstable = ["Name", "Status", "Ips", "Source", "Plan", "Profile"]
    for vm in _list:
        if field is not None:
            name = next(iter(vm))
            vminfo = [name, vm[name]]
            vmstable.append(vminfo)
            continue
        name = vm.get('name')
        status = vm.get('status')
        ip = vm.get('ip', '')
        source = vm.get('image', '')
        plan = vm.get('plan', '')
        profile = vm.get('profile', '')
        vminfo = [name, status, ip, source, plan, profile]
        vmstable.append(vminfo)
    return vmstable


mcp = FastMCP("kcli-core")


@mcp.prompt()
def prompt() -> str:
    """Indicates contexts of questions related to kcli"""
    return """You are a helpful assistant who knows everything about kcli, a powerful client and library written
    in Python and meant to interact with different virtualization providers, easily deploy and customize VMs or
    full kubernetes/OpenShift clusters. All information about kcli is available at
    https://github.com/karmab/kcli/blob/main/docs/index.md"""


@mcp.resource("resource://kcli-doc.md")
def get_doc() -> str:
    """Provides kcli documentation"""
    url = 'https://raw.githubusercontent.com/karmab/kcli/refs/heads/main/docs/index.md'
    return urlopen(url).read().decode('utf-8')


@mcp.tool()
def create_kube(cluster: str, kubetype: str = 'generic', threaded: bool = False, force: bool = False,
                disks: list = [], nets: list = [], sno_vm: bool = False,
                client: str = None, debug: bool = False, region: str = None,
                zone: str = None, namespace: str = None, overrides: dict = {}) -> dict:
    """Create cluster"""
    if disks:
        overrides['disk_size'] = disks[0]['size'] if isinstance(disks[0], dict) else disks[0]
        if len(disks) > 1:
            overrides['extra_disks'] = disks[1:]
    if nets:
        overrides['network'] = nets[0]['name'] if isinstance(nets[0], dict) else nets[0]
        if len(nets) > 1:
            overrides['extra_networks'] = nets[1:]
    if threaded:
        overrides['threaded'] = threaded
    master_parameters = [key for key in overrides if 'master' in key]
    if master_parameters:
        master_parameters = ','.join(master_parameters)
        msg = f"parameters that contain master word need to be replaced with ctlplane. Found {master_parameters}"
        return {'result': 'failure', 'reason': msg}
    sno = kubetype == 'openshift-sno' or (kubetype == 'openshift' and 'sno' in overrides and overrides['sno'])
    offline = sno and not sno_vm and (client == 'fake' or common.need_fake())
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace, offline=offline)
    if force:
        overrides['kubetype'] = kubetype
        config.delete_kube(cluster, overrides=overrides)
    confpool = overrides.get('namepool') or overrides.get('confpool')
    if cluster is None and confpool is not None:
        cluster = config.get_name_from_confpool(confpool)
    clusterprofile = overrides.get('clusterprofile')
    if clusterprofile is not None:
        if clusterprofile not in config.clusterprofiles:
            msg = f"Clusterprofile {clusterprofile} not found"
            return {'result': 'failure', 'reason': msg}
        else:
            initial_apps = overrides.get('apps', [])
            clusterprofile = config.clusterprofiles[clusterprofile]
            clusterprofiles_apps = clusterprofile.get('apps', [])
            clusterprofile.update(overrides)
            overrides = clusterprofile
            overrides['apps'] = list(dict.fromkeys(clusterprofiles_apps + initial_apps))
    return config.create_kube(cluster, kubetype, overrides=overrides)


@mcp.tool()
def create_network(name: str, cidr: str, domain: str = None, plan: str = 'kvirt', dual_cidr: str = None,
                   dual_name: str = None, dhcp: bool = True, nodhcp: bool = False, isolated: bool = False,
                   client: str = None, debug: bool = False, region: str = None,
                   zone: str = None, namespace: str = None, overrides: dict = {}) -> dict:
    """Create network"""
    nodhcp = not dhcp if dhcp is not None else nodhcp
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    k = config.k
    nat = not isolated
    dhcp = not nodhcp
    if dual_cidr is not None:
        overrides['dual_cidr'] = dual_cidr
    if dual_name is not None:
        overrides['dual_name'] = dual_name
    return k.create_network(name=name, cidr=cidr, dhcp=dhcp, nat=nat, domain=domain, overrides=overrides, plan=plan)


@mcp.tool()
def create_plan(plan: str = None, ansible: bool = False, url: str = None, path: str = None, container: bool = False,
                threaded: bool = False, inputfile: str = 'kcli_plan.yml', force: bool = False,
                client: str = None, debug: bool = False, region: str = None,
                zone: str = None, namespace: str = None, overrides: dict = {}) -> dict:
    """Create plan"""
    # if container_mode():
    #    inputfile = f"/workdir/{inputfile}"
    if 'minimum_version' in overrides:
        minimum_version = overrides['minimum_version']
        current_version = get_git_version()[0]
        if current_version != 'N/A':
            if not compare_git_versions(minimum_version, current_version):
                msg = f"Current kcli version {current_version} lower than plan minimum version {minimum_version}"
                return {'result': 'failure', 'reason': msg}
    offline = overrides.get('offline', False)
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace, offline=offline)
    _type = config.ini[config.client].get('type', 'kvm')
    overrides.update({'type': _type})
    if plan is None:
        plan = get_random_name()
    if force:
        if plan is None:
            msg = "Force requires specifying a plan name"
            return {'result': 'failure', 'reason': msg}
        else:
            config.delete_plan(plan, unregister=config.rhnunregister)
    return config.plan(plan, ansible=ansible, url=url, path=path, container=container, inputfile=inputfile,
                       overrides=overrides, threaded=threaded)


@mcp.tool()
def create_pool(pool: str, pooltype: str = 'dir', path: str = None, thinpool: bool = False,
                client: str = None, debug: bool = False, region: str = None,
                zone: str = None, namespace: str = None) -> dict:
    """Create pool"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    k = config.k
    if path is None:
        msg = "Missing path"
        return {'result': 'failure', 'reason': msg}
    return k.create_pool(name=pool, poolpath=path, pooltype=pooltype, thinpool=thinpool)


@mcp.tool()
def create_vm(name: str, profile: str,
              client: str = None, debug: bool = False, region: str = None,
              zone: str = None, namespace: str = None, overrides: dict = {}) -> dict:
    """Create vm"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    return config.create_vm(name, profile, overrides=overrides)


@mcp.tool()
def delete_kube(cluster: str, allclusters: bool = False,
                client: str = None, debug: bool = False, region: str = None,
                zone: str = None, namespace: str = None, overrides: dict = {}):
    """Delete cluster"""
    if client is not None:
        overrides['client'] = client
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    clusters = [c for c in config.list_kubes()] if allclusters else cluster
    for cluster in clusters:
        config.delete_kube(cluster, overrides=overrides)


@mcp.tool()
def delete_network(names: list = [], force: bool = False,
                   client: str = None, debug: bool = False, region: str = None,
                   zone: str = None, namespace: str = None) -> dict:
    """Delete network"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    k = config.k
    for name in names:
        result = k.delete_network(name=name, force=force)
        if result['result'] != 'success':
            return result
    return {'result': 'success'}


@mcp.tool()
def delete_plan(plans: list = [], allplans: bool = False,
                client: str = None, debug: bool = False, region: str = None,
                zone: str = None, namespace: str = None) -> dict:
    """Delete plan"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    plans = [p[0] for p in config.list_plans()] if allplans else plans
    for plan in plans:
        result = config.delete_plan(plan, unregister=config.rhnunregister)
        if 'result' in result and result['result'] != 'success':
            return result
    return {'result': 'success'}


@mcp.tool()
def delete_pool(pool: str, full: bool = False,
                client: str = None, debug: bool = False, region: str = None,
                zone: str = None, namespace: str = None) -> dict:
    """Delete pool"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    k = config.k
    return k.delete_pool(name=pool, full=full)


@mcp.tool()
def delete_vm(vm: str,
              client: str = None, debug: bool = False, region: str = None,
              zone: str = None, namespace: str = None) -> dict:
    """Delete vm"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    return config.k.delete(vm)


@mcp.tool()
def info_vm(name: str,
            client: str = None, debug: bool = False, region: str = None,
            zone: str = None, namespace: str = None) -> dict:
    """Get info of vm"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    return config.k.info(name)


@mcp.tool()
def list_clients(client: str = None, debug: bool = False) -> list:
    """List clients"""
    clientstable = ["Client", "Type", "Enabled", "Current"]
    baseconfig = Kbaseconfig(client=client, debug=debug)
    for client in sorted(baseconfig.clients):
        enabled = baseconfig.ini[client].get('enabled', True)
        _type = baseconfig.ini[client].get('type', 'kvm')
        if client == baseconfig.client:
            clientstable.append([client, _type, enabled, 'X'])
        else:
            clientstable.append([client, _type, enabled, ''])
    return clientstable


@mcp.tool()
def list_clusters(client: str = None, debug: bool = False, region: str = None,
                  zone: str = None, namespace: str = None) -> list:
    """List clusters"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    return config.k.list_kubes()


@mcp.tool()
def list_images(client: str = None, debug: bool = False, region: str = None,
                zone: str = None, namespace: str = None) -> list:
    """List images"""
    return Kconfig(client=client).k.volumes()


@mcp.tool()
def list_networks(client: str = None, debug: bool = False, region: str = None,
                  zone: str = None, namespace: str = None) -> dict:
    """List networks"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    return config.k.list_networks()


@mcp.tool()
def list_pools(client: str = None, debug: bool = False, region: str = None,
               zone: str = None, namespace: str = None) -> list:
    """List pools"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    return config.k.list_pools()


@mcp.tool()
def list_vms(client: str = None, debug: bool = False, region: str = None,
             zone: str = None, namespace: str = None) -> list:
    """List vms"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    return config.k.list()


@mcp.tool()
def scale_kube(cluster: str, kubetype: str = 'generic', ctlplanes: Optional[int] = None, workers: Optional[int] = None,
               client: str = None, debug: bool = False, region: str = None,
               zone: str = None, namespace: str = None, overrides: dict = {}) -> dict:
    """Scale cluster"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    if ctlplanes is not None:
        overrides['ctlplanes'] = ctlplanes
    if workers is not None:
        overrides['workers'] = workers
    return config.scale_kube(cluster, kubetype, overrides=overrides)


@mcp.tool()
def start_plan(plans: list = [],
               client: str = None, debug: bool = False, region: str = None,
               zone: str = None, namespace: str = None):
    """Start plan"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    for plan in plans:
        result = config.start_plan(plan)
        if 'result' in result and result['result'] != 'success':
            return result
    return {'result': 'success'}


@mcp.tool()
def start_vm(name: str,
             client: str = None, debug: bool = False, region: str = None,
             zone: str = None, namespace: str = None) -> dict:
    """Start vm"""
    return Kconfig(client).k.start(name)


@mcp.tool()
def stop_plan(plans: list = [], soft: bool = False,
              client: str = None, debug: bool = False, region: str = None,
              zone: str = None, namespace: str = None):
    """Stop plan"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    for plan in plans:
        result = config.stop_plan(plan, soft=soft)
        if 'result' in result and result['result'] != 'success':
            return result
    return {'result': 'success'}


@mcp.tool()
def stop_vm(name: str,
            client: str = None, debug: bool = False, region: str = None,
            zone: str = None, namespace: str = None) -> dict:
    """Stop vm"""
    config = Kconfig(client=client, debug=debug, region=region, zone=zone, namespace=namespace)
    return config.k.stop(name)


def main():
    mcp.run(transport="stdio")


if __name__ == "__main__":
    main()



from google.cloud import container_v1
from kvirt.common import error, success, info2, pprint, fix_typos
import os
from time import sleep
import sys
from yaml import safe_dump, safe_load


def _wait_for_operation(client, location):
    location = location.replace('https://container.googleapis.com/v1/', '')
    done = False
    timeout = 0
    while not done:
        if timeout > 60:
            return
        request = container_v1.GetOperationRequest(name=location)
        operation = client.get_operation(request=request)
        if str(operation.status) == 'Status.DONE':
            done = True
        else:
            pprint("Waiting 5s for operation to complete")
            sleep(5)
            timeout += 5
        if operation.error.message != '':
            error(f"Got Error {operation.error.message}")
            break
    return


def get_kubeconfig(config, cluster, zonal=True):
    plandir = os.path.dirname(project_init.__code__.co_filename)
    project, region, zone = project_init(config)
    client = container_v1.ClusterManagerClient()
    request = {"name": f"projects/{project}/locations/{zone if zonal else region}/clusters/{cluster}"}
    response = client.get_cluster(request=request)
    endpoint = response.endpoint
    if endpoint == '':
        return
    ca_cert = response.master_auth.cluster_ca_certificate
    rendered = config.process_inputfile(cluster, f"{plandir}/kubeconfig.j2", overrides={'endpoint': endpoint,
                                                                                        'ca_cert': ca_cert,
                                                                                        'client': config.client})
    clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
    with open(f"{clusterdir}/auth/kubeconfig", 'w') as f:
        f.write(rendered)


def project_init(config):
    if config.type != 'gcp':
        error("This is only available for gcp provider")
        sys.exit(1)
    credentials = config.options.get('credentials')
    if credentials is not None:
        os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = os.path.expanduser(credentials)
    elif 'GOOGLE_APPLICATION_CREDENTIALS' not in os.environ:
        error("set GOOGLE_APPLICATION_CREDENTIALS variable.Leaving...")
        sys.exit(1)
    project = config.options.get('project')
    if project is None:
        error("Missing project in the configuration. Leaving")
        sys.exit(1)
    region = config.options.get('region')
    zone = config.options.get('zone')
    if region is None and zone is None:
        error("Missing project in the configuration. Leaving")
        sys.exit(1)
    elif region is not None and zone is None:
        zone = f'{region}-b'
    elif region is None and zone is not None:
        region = zone[:-2]
    return project, region, zone


def scale(config, plandir, cluster, overrides):
    data = {'workers': 2}
    data.update(overrides)
    workers = data['workers']
    cluster = overrides.get('cluster', cluster or 'mygke')
    project, region, zone = project_init(config)
    client = container_v1.ClusterManagerClient()
    zonal = overrides.get('zonal', True)
    nodepool = f"projects/{project}/locations/{zone if zonal else region}/clusters/{cluster}/nodepools/{cluster}"
    request = container_v1.SetNodePoolSizeRequest(name=nodepool, node_count=workers)
    operation = client.set_node_pool_size(request=request)
    _wait_for_operation(client, operation.self_link)
    return {'result': 'success'}


def create(config, plandir, cluster, overrides, dnsconfig=None):
    data = safe_load(open(f'{plandir}/kcli_default.yml'))
    data.update(overrides)
    fix_typos(data)
    workers = data['workers']
    zonal = data['zonal']
    autopilot = data['autopilot']
    autoscaling = data['autoscaling']
    beta_apis = data['beta_apis']
    spot = data['spot']
    secureboot = data['secureboot']
    confidential = data['confidential']
    preemptible = data['preemptible']
    disk_type = data['disk_type']
    plan = cluster
    clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
    if os.path.exists(clusterdir):
        msg = f"Remove existing directory {clusterdir} or use --force"
        return {'result': 'failure', 'reason': msg}
    else:
        os.makedirs(clusterdir)
        os.mkdir(f"{clusterdir}/auth")
        with open(f"{clusterdir}/kcli_parameters.yml", 'w') as p:
            installparam = overrides.copy()
            installparam['plan'] = plan
            installparam['cluster'] = cluster
            installparam['kubetype'] = 'gke'
            installparam['zonal'] = zonal
            installparam['client'] = config.client
            safe_dump(installparam, p, default_flow_style=False, encoding='utf-8', allow_unicode=True)
    project, region, zone = project_init(config)
    clusterspec = {'name': cluster, 'enable_kubernetes_alpha': data['alpha']}
    clusterspec['resource_labels'] = {'plan': cluster, 'kube': cluster, 'kubetype': 'gke'}
    network = data['network']
    create_subnetwork = False
    networks = config.k.list_networks()
    if network != 'default':
        subnets = config.k.list_subnets()
        if network in networks:
            clusterspec['network'] = network
            legacy = networks[network]['cidr'] != ''
        elif network in subnets:
            subnet_data = config.k.info_subnet(network)
            clusterspec['network'] = subnet_data['network']
            clusterspec['subnetwork'] = network
            legacy = 'secondary_cidrs' not in subnet_data
        else:
            msg = f'Invalid subnet {network}'
            return {'result': 'failure', 'reason': msg}
    elif 'default' in networks:
        clusterspec['network'] = network
        legacy = networks[network]['cidr'] != ''
    else:
        msg = f'Invalid network {network}'
        return {'result': 'failure', 'reason': msg}
    native = data['native']
    cluster_network_ipv4, service_network_ipv4 = data['cluster_network_ipv4'], data['service_network_ipv4']
    if native:
        ip_allocation_policy = {'use_ip_aliases': True,
                                'create_subnetwork': create_subnetwork,
                                'cluster_ipv4_cidr_block': cluster_network_ipv4,
                                'services_ipv4_cidr_block': service_network_ipv4}
        clusterspec['ip_allocation_policy'] = ip_allocation_policy
    elif legacy:
        clusterspec['ip_allocation_policy'] = {'use_ip_aliases': False}
    if 'version' in overrides:
        clusterspec['initial_cluster_version'] = overrides['version']
    if beta_apis:
        clusterspec['enable_k8s_beta_apis'] = True
    if autopilot:
        clusterspec['autopilot'] = True
    nodepool = {'name': cluster, 'initial_node_count': workers}
    worker_version = data['worker_version']
    if worker_version is not None:
        nodepool['version'] = overrides['worker_version']
    if autoscaling:
        min_node_count = data['autoscaling_minimum']
        max_node_count = data['autoscaling_maximum']
        nodepool['autoscaling'] = {'enable_node_autoprovisioning': True, 'min_node_count': min_node_count,
                                   'max_node_count': max_node_count}
    nodepool_config = {'preemptible': preemptible, 'spot': spot, 'disk_type': disk_type}
    integrity_monitoring = data['integrity_monitoring']
    if secureboot or integrity_monitoring:
        nodepool_config['shielded_instance_config'] = {'enable_secure_boot': secureboot,
                                                       'enable_integrity_monitoring': integrity_monitoring}
    if confidential:
        nodepool_config['confidential_nodes'] = {'enabled': True}
    flavor = data['flavor']
    if flavor is not None:
        nodepool_config['machine_type'] = flavor
    disk_size = data['disk_size']
    if disk_size is not None:
        nodepool_config['disk_size_gb'] = disk_size
    image = data.get('image') or data.get('image_type')
    if image is not None:
        nodepool_config['image_type'] = image.upper()
    local_ssd_count = data['local_ssd_count']
    if local_ssd_count is not None:
        nodepool_config['local_ssd_count'] = local_ssd_count
    nodepool['config'] = nodepool_config
    clusterspec['node_pools'] = [nodepool]
    client = container_v1.ClusterManagerClient()
    parent = f'projects/{project}/locations/{zone if zonal else region}'
    request = container_v1.CreateClusterRequest(parent=parent, cluster=clusterspec)
    if config.debug:
        print(clusterspec)
    operation = client.create_cluster(request=request)
    if config.debug:
        print(operation)
    _wait_for_operation(client, operation.self_link)
    kubeconfig = f'{clusterdir}/auth/kubeconfig'
    while not os.path.exists(kubeconfig):
        get_kubeconfig(config, cluster, zonal=zonal)
        sleep(5)
    success(f"Kubernetes cluster {cluster} deployed!!!")
    info2(f"export KUBECONFIG=$HOME/.kcli/clusters/{cluster}/auth/kubeconfig")
    info2("export PATH=$PWD:$PATH")
    return {'result': 'success'}


def delete(config, cluster, zonal=True):
    project, region, zone = project_init(config)
    client = container_v1.ClusterManagerClient()
    cluster = f"projects/{project}/locations/{zone if zonal else region}/clusters/{cluster}"
    request = container_v1.DeleteClusterRequest(name=cluster)
    try:
        operation = client.delete_cluster(request=request)
    except Exception as e:
        error(f"Hit Issue when getting {cluster}: {e}")
        return {'result': 'failure', 'reason': e}
    _wait_for_operation(client, operation.self_link)
    return {'result': 'success'}


def list(config):
    results = {}
    project, region, zone = project_init(config)
    client = container_v1.ClusterManagerClient()
    parent = f"projects/{project}/locations/-/clusters"
    request = container_v1.ListClustersRequest(parent=parent)
    clusters = client.list_clusters(request=request).clusters
    for cluster in clusters:
        results[cluster.name] = {'type': 'gke', 'plan': None, 'vms': []}
    return results


def info(config, cluster, debug=False):
    project, region, zone = project_init(config)
    client = container_v1.ClusterManagerClient()
    clusterdir = os.path.expanduser(f'~/.kcli/clusters/{cluster}')
    kubeconfig = f'{clusterdir}/auth/kubeconfig'
    cluster = f"projects/{project}/locations/{zone}/clusters/{cluster}"
    request = container_v1.GetClusterRequest(name=cluster)
    try:
        clusterinfo = client.get_cluster(request=request)
    except Exception as e:
        error(f"Hit Issue when getting {cluster}: {e}")
        return {}
    if debug:
        print(clusterinfo)
    if os.path.exists(kubeconfig):
        results = config.info_specific_kube(cluster)
    else:
        # nodes = [f'node-{index}' for index in range(clusterinfo.current_node_count)]
        results = {'nodes': [], 'version': clusterinfo.node_pools[0].version}
    return results


def info_service(config, zonal=True):
    project, region, zone = project_init(config)
    client = container_v1.ClusterManagerClient()
    name = f"projects/{project}/locations/{zone if zonal else region}"
    request = container_v1.GetServerConfigRequest(name=name)
    response = client.get_server_config(request=request)
    default_image_type = response.default_image_type
    print(f"default_image_type: {default_image_type}")
    valid_image_types = response.valid_image_types
    print(f"valid_image_types: {valid_image_types}")
    default_cluster_version = response.default_cluster_version
    print(f"default_cluster_version: {default_cluster_version}")
    valid_master_versions = response.valid_master_versions
    print(f"valid_cluster_versions: {valid_master_versions}")
    valid_node_versions = response.valid_node_versions
    print(f"valid_worker_versions: {valid_node_versions}")
    return {}



# coding=utf-8

from kvirt.defaults import FAKECERT
from kvirt.bottle import Bottle, request, response, jinja2_view, server_names, ServerAdapter, auth_basic
from kvirt.common import pprint, error
from kvirt.baseconfig import Kbaseconfig
from kvirt.config import Kconfig
import os
import subprocess
from datetime import datetime
import functools
from tempfile import NamedTemporaryFile


basedir = f"{os.path.dirname(Bottle.run.__code__.co_filename)}/ksushy"
view = functools.partial(jinja2_view, template_lookup=[f"{basedir}/templates"])

default_user = os.environ.get('KSUSHY_USER')
default_password = os.environ.get('KSUSHY_PASSWORD')


def credentials(user, password):
    if default_user is None or default_password is None:
        return True
    elif user is None or password is None:
        return False
    elif user == default_user and password == default_password:
        return True
    else:
        return False


class SSLCherryPy(ServerAdapter):
    def run(self, handler):
        cert = NamedTemporaryFile(delete=False)
        cert.write(FAKECERT.encode())
        cert.close()
        from cheroot.ssl.pyopenssl import pyOpenSSLAdapter
        from cheroot import wsgi
        server = wsgi.Server((self.host, self.port), handler, request_queue_size=32)
        self.srv = server
        server.ssl_adapter = pyOpenSSLAdapter(cert.name, cert.name)
        try:
            server.start()
        finally:
            server.stop()
            os.unlink(cert.name)


class Ksushy():

    def __init__(self):
        app = Bottle()

        @app.hook('after_request')
        def wsgilog():
            if 'KSUSHY_SSL' not in os.environ:
                return
            ip = request.environ.get('REMOTE_ADDR')
            time = datetime.now().strftime('%d/%m/%Y %H:%M:%S')
            method = request.environ.get('REQUEST_METHOD')
            uri = request.environ.get('REQUEST_URI')
            protocol = request.environ.get('SERVER_PROTOCOL')
            status = response.status_code
            length = len(response.body)
            print(f'{ip} - - [{time}] "{method} {uri} {protocol}" {status} {length}')

        @app.route('/redfish/v1')
        @app.route('/redfish/v1/')
        @view('root.json')
        def root_resource():
            return {}

        @app.route('/redfish/v1/Managers')
        @auth_basic(credentials)
        @view('managers.json')
        def manager_collection_resource():
            return {}

        @app.route('/redfish/v1/Systems')
        @auth_basic(credentials)
        @view('systems.json')
        def system_collection_resource():
            clients = []
            baseconfig = Kbaseconfig()
            for client in baseconfig.clients:
                clients.append({"@odata.id": f"/redfish/v1/Systems/{client}"})
            return {'vms': clients, 'count': len(clients)}

        @app.route('/redfish/v1/Systems/<client>')
        @auth_basic(credentials)
        @view('systems.json')
        def system_collection_client_resource(client):
            baseconfig = Kbaseconfig()
            if client not in baseconfig.clients:
                response.status = 404
                msg = f'Client {client} not found'
                error(msg)
                return msg
            config = Kconfig(client)
            k = config.k
            vms = []
            for vm in k.list():
                vms.append({"@odata.id": f"/redfish/v1/Systems/{client}/{vm['name']}"})
            return {'vms': vms, 'count': len(vms)}

        @app.route('/redfish/v1/Systems/<client>/<name>')
        @auth_basic(credentials)
        @view('system.json')
        def system_resource_get(client, name):
            baseconfig = Kbaseconfig()
            if client not in baseconfig.clients:
                response.status = 404
                msg = f'Client {client} not found'
                error(msg)
                return msg
            config = Kconfig(client)
            k = config.k
            info = k.info(name)
            if not info:
                response.status = 404
                msg = f'VM {name} not found'
                error(msg)
                return msg
            status = 'On' if info['status'] == 'up' else 'Off'
            data = {'client': client, 'name': name, 'status': status, 'memory': info['memory'], 'cpus': info['numcpus'],
                    'virt_type': config.type}
            if 'id' in info:
                data['uuid'] = info['id']
            return data

        @app.route('/redfish/v1/Systems/<client>/<name>', method='PATCH')
        @auth_basic(credentials)
        def system_resource(client, name):
            boot = request.json.get('Boot', {})
            if not boot:
                response.status = 400
                msg = 'PATCH only works for Boot'
                error(msg)
                return msg
            target = boot.get('BootSourceOverrideTarget')
            mode = boot.get('BootSourceOverrideMode')
            if not target and not mode:
                response.status = 400
                msg = 'Missing the BootSourceOverrideTarget and/or BootSourceOverrideMode element'
                error(msg)
                return msg
            else:
                response.status = 204
                return ''

        @app.route('/redfish/v1/Systems/<client>/<name>/EthernetInterfaces')
        @auth_basic(credentials)
        @view('interfaces.json')
        def manage_interfaces(client, name):
            baseconfig = Kbaseconfig()
            if client not in baseconfig.clients:
                response.status = 404
                msg = f'Client {client} not found'
                error(msg)
                return msg
            config = Kconfig(client)
            k = config.k
            info = k.info(name)
            if not info:
                response.status = 404
                msg = f'VM {name} not found'
                error(msg)
                return msg
            macs = []
            for nic in info.get('nets', []):
                mac = nic['mac']
                macs.append({"@odata.id": f"/redfish/v1/Systems/{client}/{name}/EthernetInterfaces/{mac}"})
            return {'client': client, 'name': name, 'macs': macs, 'count': len(macs)}

        @app.route('/redfish/v1/Systems/<client>/<name>/EthernetInterfaces/<mac>')
        @auth_basic(credentials)
        @view('interface.json')
        def manage_interface(client, name, mac):
            return {'client': client, 'name': name, 'mac': mac}

        @app.route('/redfish/v1/Managers/<client>/<name>')
        @auth_basic(credentials)
        @view('manager.json')
        def manager_resource(client, name):
            return {'client': client, 'name': name, 'date_time': datetime.now().strftime('%Y-%M-%dT%H:%M:%S+00:00')}

        @app.route('/redfish/v1/Systems/<client>/<name>/Actions/ComputerSystem.Reset', method='POST')
        @auth_basic(credentials)
        def system_reset_action(client, name):
            baseconfig = Kbaseconfig()
            if client not in baseconfig.clients:
                response.status = 404
                msg = f'Client {client} not found'
                error(msg)
                return msg
            config = Kconfig(client)
            k = config.k
            reset_type = request.json.get('ResetType', 'On')
            if reset_type in ['On', 'ForceRestart']:
                try:
                    pprint(f"Starting vm {name}")
                    if config.type == 'kvm' and self.bootonce:
                        pprint(f"Forcing boot from cdrom for vm {name}")
                        k.force_cdrom(name)
                    k.start(name)
                except subprocess.CalledProcessError as e:
                    error(e)
                    response.status = 400
                    return 'Failed to poweron the server'
            else:
                try:
                    pprint(f"Stopping vm {name}")
                    k.stop(name)
                except subprocess.CalledProcessError as e:
                    error(e)
                    response.status = 400
                    return 'Failed to poweroff the server'
            response.status = 204
            return ''

        @app.route('/redfish/v1/Managers/<client>/<name>/VirtualMedia')
        @auth_basic(credentials)
        @view('virtualmedias.json')
        def virtualmedia_collection_resource(client, name):
            return {'client': client, 'name': name}

        @app.route('/redfish/v1/Managers/<client>/<name>/VirtualMedia/Cd')
        @auth_basic(credentials)
        @view('virtualmedia_cd.json')
        def virtualmedia_cd_resource(client, name):
            baseconfig = Kbaseconfig()
            if client not in baseconfig.clients:
                response.status = 404
                msg = f'Client {client} not found'
                error(msg)
                return msg
            config = Kconfig(client)
            info = config.k.info(name)
            if not info:
                response.status = 404
                msg = f'VM {name} not found'
                error(msg)
                return msg
            inserted, image_url = False, ''
            if 'iso' in info:
                inserted = True
                image_url = os.path.basename(info['iso'])
            return {'client': client, 'name': name, 'inserted': inserted, 'image_url': image_url}

        @app.route('/redfish/v1/Managers/<client>/<name>/VirtualMedia/Cd/Actions/VirtualMedia.InsertMedia',
                   method='POST')
        @auth_basic(credentials)
        def virtualmedia_insert(client, name):
            baseconfig = Kbaseconfig()
            if client not in baseconfig.clients:
                response.status = 404
                msg = f'Client {client} not found'
                error(msg)
                return msg
            config = Kconfig(client)
            if not config.k.exists(name):
                response.status = 404
                msg = f'VM {name} not found'
                error(msg)
                return msg
            image = request.json.get('Image')
            if image is None:
                response.status = 400
                msg = 'POST only works for Image'
                error(msg)
                return msg
            try:
                pprint(f"Setting iso of vm {name} to {image}")
                info = config.k.info(name)
                if 'redfish_iso' in info:
                    iso = info['redfish_iso']
                else:
                    iso = os.path.basename(image)
                    token_iso = os.path.basename(image).split('?')[0]
                    if token_iso != iso:
                        iso = f"boot-{token_iso}.iso"
                isos = [os.path.basename(i) for i in config.k.volumes(iso=True)]
                if iso not in isos:
                    result = config.download_image(pool=config.pool, image=iso, url=image)
                    if result['result'] != 'success':
                        raise Exception(result['reason'])
                config.update_vm(name, {'iso': iso})
            except Exception as e:
                msg = f'Failed to mount virtualcd. Hit {e}'
                error(msg)
                response.status = 500
                return msg
            response.status = 204
            return ''

        @app.route('/redfish/v1/Managers/<client>/<name>/VirtualMedia/Cd/Actions/VirtualMedia.EjectMedia',
                   method='POST')
        @auth_basic(credentials)
        def virtualmedia_eject(client, name):
            baseconfig = Kbaseconfig()
            if client not in baseconfig.clients:
                response.status = 404
                msg = f'Client {client} not found'
                error(msg)
                return msg
            config = Kconfig(client)
            if not config.k.exists(name):
                response.status = 404
                msg = f'VM {name} not found'
                error(msg)
                return msg
            try:
                pprint(f"Setting iso of vm {name} to None")
                info = config.k.info(name)
                if 'iso' in info:
                    config.update_vm(name, {'iso': None})
            except subprocess.CalledProcessError as e:
                error(e)
                return ('Failed to unmount virtualcd', 400)
            response.status = 204
            return ''

        @app.route('/redfish/v1/Systems/<client>/<name>/BIOS')
        @auth_basic(credentials)
        @view('bios.json')
        def bios_resource(client, name):
            return {'client': client, 'name': name}

        self.app = app
        self.port = os.environ.get('KSUSHY_PORT', 9000)
        self.debug = 'KSUSHY_DEBUG' in os.environ
        self.ipv6 = 'KSUSHY_IPV6' in os.environ
        self.host = '::' if self.ipv6 else '0.0.0.0'
        self.bootonce = 'KSUSHY_BOOTONCE' in os.environ

    def run(self):
        data = {'host': self.host, 'port': self.port, 'debug': self.debug}
        if 'KSUSHY_SSL' in os.environ:
            server_names['sslcherrypy'] = SSLCherryPy
            data['server'] = 'sslcherrypy'
        self.app.run(**data)



# -*- coding: utf-8 -*-

from datetime import datetime
from kvirt import common
from kvirt.common import error, pprint, warning, sdn_ip
from kvirt.defaults import METADATA_FIELDS, UBUNTUS
from kvirt.providers.sampleprovider import Kbase
import os
from pathlib import Path
import proxmoxer
from proxmoxer.tools import Tasks
import re
from subprocess import CalledProcessError, call, check_call
import time
from textwrap import dedent
from tempfile import TemporaryDirectory
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
import webbrowser
from yaml import safe_dump


VM_STATUS = {"running": "up", "stopped": "down", "unknown": "unk"}


def patch_cmds(image, cmds):
    gcmds = []
    lower = image.lower()
    if lower.startswith('fedora') or lower.startswith('rhel') or lower.startswith('centos'):
        gcmds.append('yum -y install qemu-guest-agent')
        gcmds.append('systemctl enable --now qemu-guest-agent')
    elif lower.startswith('debian') or [x for x in UBUNTUS if x in lower] or 'ubuntu' in lower:
        gcmds.append('apt-get update')
        gcmds.append('apt-get -y install qemu-guest-agent')
        gcmds.append('/etc/init.d/qemu-guest-agent start')
        gcmds.append('update-rc.d qemu-guest-agent defaults')
    index = 1 if cmds and 'sleep' in cmds[0] else 0
    if image.startswith('rhel'):
        subindex = [i for i, value in enumerate(cmds) if value.startswith('subscription-manager')]
        if subindex:
            index = subindex.pop() + 1
    return cmds[:index] + gcmds + cmds[index:]


def sizeof_fmt(num, suffix="B"):
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


class Kproxmox(Kbase):
    def __init__(
        self,
        host="127.0.0.1",
        port=None,
        user="root@pam",
        password=None,
        token_name=None,
        token_secret=None,
        filtertag=None,
        node=None,
        verify_ssl=True,
        imagepool=None,
        debug=False,
    ):
        self.conn = proxmoxer.ProxmoxAPI(
            host,
            port=port,
            user=user,
            password=password,
            token_name=token_name,
            token_value=token_secret,
            verify_ssl=verify_ssl,
            timeout=60,
        )
        self.filtertag = filtertag
        if not node:
            # Get responding node
            status = self.conn.cluster.status.get()
            for s in status:
                if s["type"] == "node" and s["local"] == 1:
                    node = s["name"]
                    break
        self.node = node
        self.imagepool = imagepool
        self.host = host
        self.user = user
        self.debug = debug

    def info_host(self):
        nodes = self.conn.cluster.resources.get(type="node")
        cl_status = self.conn.cluster.status.get()
        data = {
            "cluster_name": "",
            "nodes_online": 0,
            "cpus": 0,
            "memory_used": 0,
            "memory_total": 0,
            "vms_running": 0,
            "storages": [],
        }

        for status in cl_status:
            if status["type"] != "cluster":
                data["nodes_online"] += status["online"]
                continue
            data["cluster_name"] = status["name"]

        for node in nodes:
            if node["status"] != "online":
                continue
            data["memory_total"] += node["maxmem"]
            data["memory_used"] += node["mem"]
            data["cpus"] += node["maxcpu"]

        all_vms = self.conn.cluster.resources.get(type="vm")
        data["vms_running"] = len(
            list(filter(lambda v: v["status"] == "running", all_vms))
        )

        data["storages"] = self._get_storages()

        # Human readable sizes
        data["memory_total"] = sizeof_fmt(data["memory_total"])
        data["memory_used"] = sizeof_fmt(data["memory_used"])

        return data

    def list(self):
        vms = []
        for vm in self._get_all_vms():
            name, _type, template = vm['name'], vm['type'], vm['template']
            if _type == 'lxc' or template == 1:
                continue
            try:
                vms.append(self.info(name, vm=vm))
            except:
                continue
        return sorted(vms, key=lambda x: x['name'])

    def volumes(self, iso=False):
        return [t["name"] for t in self._get_isos()] if iso else [t["name"] for t in self._get_templates()]

    def _check_node(self, node):
        # Check target node
        ret = None
        cluster_status = self.conn.cluster.status.get()
        for n in cluster_status:
            if n["type"] == "node" and n["name"] == node:
                ret = n
                break
        return ret

    def _check_storage(self, pool, node=None):
        if not node:
            node = self.node
        # Check that template storage is available on this node
        storages = self.conn.nodes(node).storage.get(content="images")
        ret = None
        for s in storages:
            if s["storage"] == pool:
                ret = s
        return ret

    def add_image(self, url, pool, cmds=[], name=None, size=None, convert=False):
        if self.imagepool:
            pool = self.imagepool

        if url.endswith(".iso") or name.endswith(".iso"):
            return self.add_iso(url, pool, name=name)

        if pool is None:
            return {"result": "failure", "reason": "No pool storage defined."}

        # Check if template already exists
        if self._get_template_info(name):
            pprint(f"Image {name} already there")
            return {"result": "success", "found": True}

        new_vmid = self.conn.cluster.nextid.get()
        shortimage = os.path.basename(url).split("?")[0]

        is_valid = False
        uncompresscmd = None
        compressed = {"bz2": "bunzip2 -f", "gz": "gunzip -f", "xz": "unxz -f", "zst": "zstd --decompress"}
        for ext in reversed(shortimage.split(".")[-2:]):
            if ext in compressed.keys():
                uncompresscmd = compressed[ext]
                continue
            if ext in ("qcow2", "img"):
                is_valid = True
            break
        if not is_valid:
            return {"result": "failure", "reason": "Image format unsupported"}

        if shortimage.endswith(".img"):
            shortimage = Path(shortimage).with_suffix(".qcow2").name

        # Check target node
        pve_node = self._check_node(self.node)
        if not pve_node:
            return {"result": "failure", "reason": f"node {self.node} not found."}

        # Check that template storage is available on this node
        if not self._check_storage(pool, node=pve_node["name"]):
            return {"result": "failure", "reason": f"storage {pool} not found on {pve_node['name']}."}

        # Download image in /var/lib/vz/images/0
        downloadpath = "/var/lib/vz/images/0"
        ssh_user = "root"

        downloadcmds = [
            f"mkdir -p {downloadpath}",
            f"curl -L '{url}' -o {downloadpath}/{shortimage}",
        ]
        if uncompresscmd:
            downloadcmds.append(f"{uncompresscmd} {downloadpath}/{shortimage}")
            shortimage = Path(shortimage).stem
        sshcmd = f'ssh {ssh_user}@{pve_node["ip"]} "{" && ".join(downloadcmds)}"'
        pprint(f"Downloading image {name} on {pve_node['name']}...")
        code = call(sshcmd, shell=True)
        if code != 0:
            return {"result": "failure", "reason": "Unable to download indicated image"}

        now = time.strftime("%d-%m-%Y %H:%M", time.gmtime())
        description = f"Image created with kcli from {shortimage} on {now}"

        # Create template
        ok, status = self._wait_for(
            self.conn.nodes(pve_node["name"]).qemu.post(
                vmid=new_vmid,
                name=name,
                scsihw="virtio-scsi-pci",
                virtio0=f"file={pool}:0,import-from=local:0/{shortimage}",
                description=dedent(description),
            )
        )
        if not ok:
            return {"result": "failure", "reason": status}

        self._wait_for(self.conn.nodes(pve_node["name"]).qemu(new_vmid).template.post())
        pprint(f"Image {name} created on {pve_node['name']}")

        # Remove image in /var/lib/vz/images/0
        pprint(f"Removing image {downloadpath}/{shortimage} on {pve_node['name']}...")
        sshcmd = f'ssh {ssh_user}@{pve_node["ip"]} "rm {downloadpath}/{shortimage}"'
        code = call(sshcmd, shell=True)
        if code != 0:
            warning("Unable to delete image.")

        # Wait until template info are up-to-date
        while not self._get_template_info(name):
            time.sleep(1)

        return {"result": "success"}

    def delete_image(self, image, pool=None):
        if image.endswith(".iso"):
            return self.delete_iso(image, pool=pool)
        # Check if template exists
        template = self._get_template_info(image)
        if template is None:
            return {"result": "failure", "reason": f"Image {image} not found"}

        ok, status = self._wait_for(
            self.conn.nodes(template["node"]).qemu(template["vmid"]).delete()
        )
        if not ok:
            return {"result": "failure", "reason": status}
        return {"result": "success"}

    def delete_iso(self, image, pool=None):
        if self.imagepool:
            pool = self.imagepool

        if pool is None:
            return {"result": "failure", "reason": "No pool storage specified."}

        isos = [i for i in self._get_isos() if i["name"] == image and i["pool"] == pool]
        if not isos:
            error(f"ISO {image} not found")
            return {"result": "failure", "reason": f"ISO {image} not found"}

        ok, status = self._wait_for(
            self.conn.nodes(self.node).storage(pool).content(f"iso/{image}").delete()
        )
        if not ok:
            return {"result": "failure", "reason": status}
        return {"result": "success"}

    def add_iso(self, url, pool, name=None):
        if self.imagepool:
            pool = self.imagepool
        if pool is None:
            return {"result": "failure", "reason": "No pool storage defined."}

        # Check if iso already exists
        isos = [i for i in self._get_isos() if i["name"] == name and i["pool"] == pool]
        if isos:
            pprint(f"ISO {name} already there")
            return {"result": "success", "found": True}

        # Check target node
        pve_node = self.node
        if not self._check_node(pve_node):
            return {"result": "failure", "reason": f"node {pve_node} not found."}

        # Check that template storage is available on this node
        if not self._check_storage(pool, node=pve_node):
            return {"result": "failure", "reason": f"storage {pool} not found on {pve_node}."}

        # All good, upload ISO
        if os.path.exists(url):
            # is path
            with open(url, "rb") as f:
                pprint(f"uploading {name} to {pve_node}...")
                ok, status = self._wait_for(
                    self.conn.nodes(pve_node)
                    .storage(pool)
                    .upload.post(content="iso", filename=f)
                )
        else:
            # is url
            ok, status = self._wait_for(
                self.conn.nodes(pve_node)
                .storage(pool)("download-url")
                .post(content="iso", filename=name, url=url)
            )
        if not ok:
            return {"result": "failure", "reason": status}
        return {"result": "success"}

    def create(
        self,
        name,
        virttype=None,
        profile="",
        flavor=None,
        plan="kvirt",
        cpumodel="host",
        cpuflags=[],
        cpupinning=[],
        numcpus=2,
        memory=512,
        guestid="guestrhel764",
        pool=None,
        image=None,
        disks=[{"size": 10}],
        disksize=10,
        diskthin=True,
        diskinterface="virtio",
        nets=[],
        iso=None,
        vnc=True,
        cloudinit=True,
        reserveip=False,
        reservedns=False,
        reservehost=False,
        start=True,
        keys=[],
        cmds=[],
        ips=None,
        netmasks=None,
        gateway=None,
        nested=True,
        dns=None,
        domain=None,
        tunnel=False,
        files=[],
        enableroot=False,
        overrides={},
        tags=[],
        storemetadata=False,
        sharedfolders=[],
        cmdline=None,
        placement=[],
        autostart=False,
        cpuhotplug=False,
        memoryhotplug=False,
        numamode=None,
        numa=[],
        pcidevices=[],
        tpm=False,
        rng=False,
        metadata={},
        securitygroups=[],
        vmuser=None,
        guestagent=True,
    ):
        imagepool = self.imagepool or pool
        if overrides.get('lxc', False):
            if image is None:
                return {"result": "failure", "reason": "Image not specified"}
            if '/' not in image:
                image = f'{imagepool}:vztmpl/{image}'
            if image not in self.list_images():
                return {"result": "failure", "reason": f"Image {image} not found"}
            disk = disks[0] if disks else [disksize]
            disksize = disk.get('size', disksize) if isinstance(disk, dict) else disk
            password = overrides.get('rootpassword', 'password')
            params = {'vmid': self.conn.cluster.nextid.get(), 'hostname': name, 'memory': memory, 'swap': 512,
                      'cores': numcpus, 'rootfs': f'{imagepool}:{disksize}', 'ostemplate': image,
                      'password': password, 'net0': 'name=eth0,bridge=vmbr0,ip=dhcp', 'start': 1}
            params['description'] = f'image={os.path.basename(image)},plan={plan},profile={profile}'
            self._wait_for(self.conn.nodes(self.node).lxc.post(**params))
            return {"result": "success"}
        qemuextra = overrides.get('qemuextra')
        if qemuextra is not None and not self.user.startswith('root'):
            return {"result": "failure", "reason": "Adjusting arg requires root user"}
        enableiommu = overrides.get('iommu', False)
        default_diskinterface = diskinterface
        needs_ignition = image is not None and (common.needs_ignition(image) or 'ignition_file' in overrides)
        if cpumodel == "host-model":
            cpumodel = "host"
        machine = None
        uefi = overrides.get('uefi', False)
        uefi_legacy = overrides.get('uefi_legacy', False)
        secureboot = overrides.get('secureboot', False)
        if (uefi or uefi_legacy or secureboot or enableiommu):
            machine = 'q35'
        # Check if vm already exists
        if self._get_vm_id(name) is not None:
            return {"result": "failure", "reason": f"VM {name} already exists"}

        # Get next available ID
        new_vmid = self.conn.cluster.nextid.get()

        if image is not None:
            # Get image template
            template = self._get_template_info(image)
            if template is None:
                return {"result": "failure", "reason": f"image {image} not found. Use kcli download image {image}."}
        if iso is not None:
            if not [i for i in self._get_isos() if i["name"] == iso and i["pool"] == imagepool]:
                return {"result": "failure", "reason": f"ISO {iso} not found"}

        # Check target node
        node = overrides.get("node") or self.node or template["node"]
        cluster_status = self.conn.cluster.status.get()
        for n in cluster_status:
            if n["type"] == "node" and n["name"] == node:
                pve_node = n
                break
        if not pve_node:
            return {"result": "failure", "reason": f"node {node} not found."}

        # Check pool storage
        storages = self.conn.nodes(node).storage.get(content="images")
        storage = None
        for s in storages:
            if s["storage"] == pool:
                storage = s
        if not storage:
            return {"result": "failure", "reason": f"storage {pool} not found on {node}."}

        if image is not None:
            # Clone template
            linked_clone = True
            if (self.imagepool and self.imagepool != pool) or overrides.get("pool"):
                linked_clone = False

            self._wait_for(
                self.conn.nodes(template["node"])
                .qemu(template["vmid"])
                .clone.post(
                    newid=new_vmid,
                    name=name,
                    target=node,
                    storage=pool if not linked_clone else None,
                    full=int(not linked_clone),
                )
            )
        else:
            # Create empty VM
            ok, status = self._wait_for(self.conn.nodes(node).qemu.post(vmid=new_vmid, name=name,
                                                                        scsihw="virtio-scsi-pci"))
            if not ok:
                return {"result": "failure", "reason": status}

        new_vm = self.conn.nodes(node).qemu(new_vmid)

        # Add tag
        if self.filtertag:
            new_vm.config.post(tags=self.filtertag)

        # Metadata
        now = time.strftime("%d-%m-%Y %H:%M", time.gmtime())
        metadata["creationdate"] = now
        description = []
        for entry in [field for field in metadata if field in METADATA_FIELDS]:
            description.append(f"{entry}={metadata[entry]}")
        description = ','.join(description)

        vm_data = {'name': name, 'cores': numcpus, 'memory': memory, 'cpu': cpumodel, 'agent': 'enabled=1',
                   'description': dedent(description), 'onboot': 1}

        sriov_nic = False
        for index, net in enumerate(nets):
            if isinstance(net, str):
                netname = net
                net = {'name': netname}
            if net.get('sriov', False):
                nets[index]['type'] = 'igb'
                nets[index]['vfio'] = True
                nets[index]['noconf'] = True
                sriov_nic = True
                if machine is None:
                    machine = 'q35'
                    warning("Forcing machine type to q35")
            nettype = net.get('type', 'virtio')
            bridge = self._get_default_network(node) if net['name'] == 'default' else net["name"]
            vm_data[f'net{index}'] = f"model={nettype},bridge={bridge}"
            mac = net.get('mac')
            if mac is not None:
                vm_data[f'net{index}'] += f",macaddr={mac}"
            vlan = net.get('vlan')
            if vlan is not None:
                if not isinstance(vlan, int):
                    return {'result': 'failure', 'reason': f"Invalid vlan value in nic {index}. Must be an int"}
                vm_data[f'net{index}'] += f",tag={vlan}"
                del nets[index]['vlan']
            multiqueues = net.get('multiqueues')
            if multiqueues is not None:
                if not isinstance(multiqueues, int):
                    return {'result': 'failure', 'reason': f"Invalid multiqueues value in nic {index}. Must be an int"}
                elif not 0 < multiqueues < 257:
                    return {'result': 'failure', 'reason': f"multiqueues value in nic {index} not between 0 and 256 "}
                else:
                    vm_data[f'net{index}'] += f",queues={multiqueues}"
            mtu = net.get('mtu')
            if mtu is not None:
                if not isinstance(mtu, int):
                    return {'result': 'failure', 'reason': f"Invalid mtu value in nic {index}. Must be an int"}
                vm_data[f'net{index}'] += f",mtu={mtu}"

        # ISO
        if iso is not None:
            vm_data['cdrom'] = f"file={imagepool}:iso/{iso},media=cdrom"

        # Cloudinit
        userdata, netdata = None, None
        if image is not None:
            node_ip = pve_node["ip"]
            noname = overrides.get('noname', False)
            meta = safe_dump({"instance-id": name, "local-hostname": name} if not noname else {})
            code = self._upload_file(node_ip, f"{name}-metadata.yaml", meta)
            if code != 0:
                return {"result": "failure", "reason": "Unable to upload metadata"}
            if needs_ignition:
                ignition = 'qemu' in image
                if not self.user.startswith('root'):
                    return {"result": "failure", "reason": "Adjusting arg requires root user"}
                ignition_path = f"{name}.ign" if ignition else f"{name}-userdata.yaml"
                version = common.ignition_version(image)
                userdata = common.ignition(name=name, keys=keys, cmds=cmds, nets=nets, gateway=gateway, dns=dns,
                                           domain=domain, files=files, enableroot=enableroot,
                                           overrides=overrides, version=version, plan=plan, image=image,
                                           vmuser=vmuser)
                code = self._upload_file(node_ip, ignition_path, userdata)
                if code != 0:
                    return {"result": "failure", "reason": "Unable to upload userdata"}
                if not ignition:
                    vm_data['citype'] = 'configdrive2'
                else:
                    arg = f" -fw_cfg name=opt/com.coreos/config,file=/var/lib/vz/snippets/{ignition_path}"
                    qemuextra = qemuextra + arg if qemuextra is not None else arg
                    userdata = None
            else:
                cmds = patch_cmds(image, cmds)
                userdata, meta, netdata = common.cloudinit(name=name, keys=keys, cmds=cmds, nets=nets,
                                                           gateway=gateway, dns=dns, domain=domain,
                                                           files=files, enableroot=enableroot, overrides=overrides,
                                                           storemetadata=storemetadata, machine='proxmox',
                                                           image=image, vmuser=vmuser)
                code = self._upload_file(node_ip, f"{name}-userdata.yaml", userdata)
                if code != 0:
                    return {"result": "failure", "reason": "Unable to upload userdata"}
                code = self._upload_file(node_ip, f"{name}-netdata.yaml", netdata)
                if code != 0:
                    return {"result": "failure", "reason": "Unable to upload netdata"}
        if userdata is not None:
            vm_data['cicustom'] = f'meta=local:snippets/{name}-metadata.yaml'
            vm_data['cicustom'] += f',user=local:snippets/{name}-userdata.yaml'
            if netdata is not None:
                vm_data['cicustom'] += f',network=local:snippets/{name}-netdata.yaml'
            vm_data['ide0'] = f"{pool}:cloudinit"

        initial_disks = self._get_current_disks(new_vm.config.get())
        # Disks
        default_disksize = disksize
        for index, disk in enumerate(disks):
            diskinterface = default_diskinterface
            diskserial = None
            diskwwn = None
            if isinstance(disk, str) and disk.isdigit():
                disksize = int(disk)
            elif isinstance(disk, int):
                disksize = disk
            elif isinstance(disk, dict):
                disksize = disk.get("size", default_disksize)
                diskinterface = disk.get('interface', default_diskinterface)
                diskwwn = disk.get('wwn')
                diskserial = disk.get('serial')
            else:
                disksize = default_disksize
            diskname = f"{diskinterface}{index}"
            if index < len(initial_disks):
                current_diskname = initial_disks[index]['name']
                current_disksize = int(initial_disks[index]['size'])
                if disksize != current_disksize and current_disksize > 1:
                    pprint(f"Waiting for image disk {index} to be resized")
                    new_vm.resize.put(disk=diskname, size=f"{disksize}G")
                vm_data[current_diskname] = initial_disks[index]['full']
            else:
                vm_data[diskname] = f"file={pool}:{disksize},format=qcow2"
                if diskwwn is not None:
                    vm_data[diskname] += ',wwn={diskwwn}'
                if diskserial is not None:
                    vm_data[diskname] += ',serial={diskserial}'

        if machine is not None:
            vm_data['machine'] = f'type={machine}'
            if enableiommu:
                vm_data['machine'] += ',viommu=intel'

        if uefi or uefi_legacy or secureboot:
            vm_data['bios'] = 'ovmf'
            if secureboot:
                vm_data['efidisk0'] = f"{pool}:32"
                vm_data['smbios1'] = 'uuid=auto'
            if tpm:
                vm_data['tpmstate'] = f"{pool}:4"
                vm_data['tpmversion'] = '2.0'

        if qemuextra is not None:
            vm_data['args'] = qemuextra

        qemuextra = overrides.get('rng', False)
        if rng:
            vm_data['rng'] = 'source=/dev/urandom'

        if sriov_nic:
            vm_data['acpi'] = 1

        for index, cell in enumerate(numa):
            cellid = cell.get('id', index)
            cellcpus = cell.get('vcpus')
            cellmemory = cell.get('memory')
            # siblings = cell.get('siblings', [])
            if cellcpus is None or cellmemory is None:
                msg = f"Can't properly use cell {index} in numa block"
                return {'result': 'failure', 'reason': msg}
            new_numa = f"cpus={cellcpus},memory={cellmemory},hostnodes={cellid}"
            if numamode is not None:
                new_numa += f',policy={numamode}'
            vm_data[f'numa{cellid}'] = new_numa

        for index, pcidevice in enumerate(pcidevices):
            vm_data['hostpci{index}'] = pcidevices

        # Configure VM
        self._wait_for(new_vm.config.post(**vm_data))

        # Start VM
        if start:
            ok, status = self._wait_for(self.conn.nodes(node).qemu(new_vmid).status.start.post())
            if not ok:
                return {"result": "failure", "reason": status}
        # Wait until vm info are up-to-date
        while self._get_vm_id(name) is None:
            time.sleep(1)
        return {"result": "success"}

    def info(self, name, output="plain", fields=[], values=False, vm=None, debug=False):
        if vm is None:
            listinfo = False
            vm_info = self._get_vm_info(name)
            if vm_info is None:
                error(f"VM {name} not found")
                return {}
        else:
            listinfo = True
            vm_info = vm
        vm = self._get_vm(name)
        vm_config = vm.config.get()
        if 'meta' not in vm_config:
            return self._get_lxc_info(name, vm_config)
        yamlinfo = {"name": vm_info["name"], "node": vm_info["node"], "id": vm_info["vmid"],
                    "status": VM_STATUS.get(vm_info["status"]), "numcpus": vm_info["maxcpu"],
                    "memory": int(vm_info["maxmem"] / 1024 / 1024), 'nets': [], 'disks': []}
        timestamp = int(self._parse_notes(vm_config['meta'])['ctime'])
        yamlinfo['creationdate'] = datetime.fromtimestamp(timestamp).strftime('%d-%m-%Y %H:%M')
        description_data = self._parse_notes(vm_config.get("description"))
        metadata = {k: description_data[k] for k in description_data if k in METADATA_FIELDS}
        yamlinfo.update(metadata)
        if 'image' in yamlinfo:
            yamlinfo['user'] = common.get_user(yamlinfo['image'])
        kubetype = yamlinfo.get('kubetype')
        cluster_network = yamlinfo.get('cluster_network')

        if vm_info["status"] == "running":
            try:
                nets = vm.agent("network-get-interfaces").get().get("result")
            except proxmoxer.core.ResourceException:
                nets = []
            ips = []
            for net in nets:
                for nic in net.get("ip-addresses", []):
                    ip = nic["ip-address"]
                    if not nic["ip-address"].startswith(tuple(['127.0.0.1', '::1', 'fe80::', '169.254.169', 'fd69']))\
                       and not sdn_ip(ip, kubetype, cluster_network):
                        ips.append(ip)
            if ips and 'ip' not in yamlinfo:
                ip4s = [i for i in ips if ':' not in i]
                ip6s = [i for i in ips if i not in ip4s]
                yamlinfo['ip'] = ip4s[0] if ip4s else ip6s[0]
            if len(ips) > 1:
                yamlinfo['ips'] = ips
        if listinfo:
            return yamlinfo
        for entry in vm_config:
            if re.findall(r"net\d+", entry):
                device = entry
                nic_data = vm_config[entry].split(',')
                network = nic_data[1].split('=')[1]
                networktype, mac = nic_data[0].split('=')
                yamlinfo['nets'].append({'device': device, 'mac': mac.lower(), 'net': network, 'type': networktype})
            if re.findall(r"(scsi|ide|virtio)\d+", entry):
                device, diskformat, drivertype = entry, entry, entry
                disk_data = vm_config[entry].split(',')
                path = disk_data[0]
                disktype, disksize = disk_data[-1].split('=')
                if disktype == 'media':
                    continue
                yamlinfo['disks'].append({'device': device, 'size': disksize, 'format': diskformat, 'type': drivertype,
                                          'path': path})
        yamlinfo['nets'] = sorted(yamlinfo['nets'], key=lambda x: x['device'])
        yamlinfo['disks'] = sorted(yamlinfo['disks'], key=lambda x: x['device'])
        if debug:
            yamlinfo['debug'] = vm_config
        return yamlinfo

    def delete(self, name, snapshots=False):
        vm_info = self._get_vm_info(name)
        if not vm_info:
            return {"result": "failure", "reason": f"VM {name} not found."}
        if vm_info["status"] == "running":
            self.stop(name, soft=False)
        vm = self.conn.nodes(vm_info["node"]).qemu(vm_info["vmid"])
        ok, status = self._wait_for(vm.delete())
        if not ok:
            return {"result": "failure", "reason": status}
        return {"result": "success"}

    def stop(self, name, soft=False):
        vm = self._get_vm(name)
        if vm is None:
            return {"result": "failure", "reason": f"VM {name} not found."}
        if soft:
            self._wait_for(vm.status.shutdown.post())
        else:
            self._wait_for(vm.status.stop.post())
        return {"result": "success"}

    def start(self, name):
        vm = self._get_vm(name)
        if vm is None:
            return {"result": "failure", "reason": f"VM {name} not found."}
        self._wait_for(vm.status.start.post())
        return {"result": "success"}

    def update_memory(self, name, memory):
        vm = self._get_vm(name)
        if vm is None:
            return {"result": "failure", "reason": f"VM {name} not found."}
        self._wait_for(vm.config.post(memory=memory))
        return {"result": "success"}

    def list_pools(self):
        return self._get_storages().keys()

    def list_networks(self):
        nets = {}
        nodes = self.conn.cluster.resources.get(type="node")
        for n in nodes:
            if n["status"] == "online":
                node_nets = self.conn.nodes(n["node"]).network.get(type="any_bridge")
                for net in node_nets:
                    if "active" in net and net["active"]:
                        netname = f"{n['node']}/{net['iface']}"
                        nets[netname] = {
                            "cidr": net.get("cidr", ""),
                            "dhcp": True,
                            "type": "bridged",
                            "mode": "N/A",
                        }

        return nets

    def list_disks(self):
        storages = self._get_storages()
        disks = {}
        for pool, pool_info in storages.items():
            node = pool_info["node"].split(",")[0]
            images = (
                self.conn.nodes(node)
                .storage(pool.split("/")[-1])
                .content.get(content="images")
            )
            for image in images:
                filename = os.path.basename(image["volid"].split(":")[-1])
                disks[filename] = {"pool": pool, "path": image["volid"]}
        return disks

    def ip(self, name):
        info = self.info(name)
        if info and "ip" in info:
            return info["ip"]
        return None

    def get_pool_path(self, pool):
        storages = self.conn.storage.get()
        node = None
        if "/" in pool:
            node, pool = pool.split("/")

        storage = None
        for s in storages:
            if s["storage"] == pool:
                storage = s
                break
        if not storage:
            return {"result": "failure", "reason": f"pool {pool} not found."}

        if "path" in storage:
            return storage["path"]
        if storage["type"] in ["lvm", "lvmthin"]:
            return f"[{storage['type']}] vg:{storage['vgname']}"

        return f"[{storage['type']}]"

    def _get_storages(self):
        all_storages = self.conn.cluster.resources.get(type="storage")

        storages = {}
        for storage in all_storages:
            if storage["status"] != "available" or "images" not in storage["content"]:
                continue
            if storage["shared"] == 0:
                storage_name = f"{storage['node']}/{storage['storage']}"
            else:
                storage_name = storage["storage"]
            if storage_name in storages:
                storages[storage_name]["node"] = ",".join(
                    storages[storage_name]["node"].split(",") + [
                        storage["node"],
                    ]
                )
            else:
                free_size = sizeof_fmt(storage["maxdisk"] - storage["disk"])
                total_size = sizeof_fmt(storage["maxdisk"])
                storages[storage_name] = {
                    "type": storage["plugintype"],
                    "node": storage["node"],
                    "size": f"{free_size} free / {total_size} total",
                }

        return storages

    def exists(self, name):
        return self._get_vm(name) is not None

    def _get_vm_info(self, name):
        all_vms = self._get_all_vms()
        for vm in all_vms:
            if "name" in vm and vm["name"] == name:
                return vm

    def _get_vm_id(self, name):
        vm_info = self._get_vm_info(name)
        return vm_info['vmid'] if vm_info is not None else None

    def _get_vm(self, name):
        for vm in self._get_all_vms():
            if vm.get("name", '') == name:
                _id, _type, _node = vm['vmid'], vm['type'], vm['node']
                return self.conn.nodes(_node).lxc(_id) if _type == 'lxc' else self.conn.nodes(_node).qemu(_id)

    def _get_all_vms(self):
        all_vms = self.conn.cluster.resources.get(type="vm")
        if self.filtertag is not None:
            return filter(lambda v: "tags" in v and self.filtertag in v["tags"], all_vms)
        else:
            return filter(lambda v: v["status"] != "unknown", all_vms)

    def _get_template_info(self, name):
        all_templates = self._get_templates()
        for template in all_templates:
            if template["name"] == name:
                return template

    def _get_templates(self):
        all_vms = self.conn.cluster.resources.get(type="vm")
        return filter(lambda v: "template" in v and v["template"] == 1, all_vms)

    def _get_isos(self):
        isos = []
        processed = []
        for node in self.conn.nodes.get():
            if node["status"] == "online":
                # list iso storage
                storages = self.conn.nodes(node["node"]).storage.get(content="iso")
                for storage in storages:
                    if storage["shared"] == 1:
                        storage_name = storage["storage"]
                    else:
                        storage_name = f"{node['node']}-{storage['storage']}"
                    if storage_name not in processed:
                        for image in (
                            self.conn.nodes(node["node"])
                            .storage(storage["storage"])
                            .content.get(content="iso")
                        ):
                            image["name"] = os.path.basename(
                                image["volid"].split(":")[-1]
                            )
                            image["pool"] = storage["storage"]
                            isos.append(image)
                        processed.append(storage_name)
        return isos

    def _get_iso_info(self, name):
        all_isos = self._get_isos()
        for iso in all_isos:
            if iso["name"] == name:
                return iso

    def _wait_for(self, task):
        ret = Tasks.blocking_status(self.conn, task, polling_interval=1)
        exitstatus = ret["exitstatus"]
        if exitstatus.endswith("OK") or exitstatus.startswith("WARNING"):
            return True, None
        return False, exitstatus

    def _parse_notes(self, notes):
        data = {}
        if notes is None:
            return data
        for entry in notes.split(','):
            desc = entry.split('=')
            if len(desc) == 2:
                field, value = desc
                data[field] = value
        return data

    def create_subnet(self, name, cidr, dhcp=True, nat=True, domain=None, plan='kvirt', overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def delete_subnet(self, name, force=False):
        print("not implemented")
        return {'result': 'success'}

    def update_subnet(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def list_dns_zones(self):
        print("not implemented")
        return []

    def list_security_groups(self, network=None):
        print("not implemented")
        return []

    def create_security_group(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def delete_security_group(self, name):
        print("not implemented")
        return {'result': 'success'}

    def update_security_group(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def console(self, name, tunnel=False, tunnelhost=None, tunnelport=22, tunneluser='root', web=False):
        vm_info = self._get_vm_info(name)
        if vm_info is None:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vm_id = os.path.basename(vm_info['id'])
        vmurl = f"https://{self.host}:8006/?console=kvm&novnc=1&vmid={vm_id}&vmname={name}&node=pve&resize=off"
        if self.debug or os.path.exists("/i_am_a_container"):
            msg = f"Open the following url:\n{vmurl}" if os.path.exists("/i_am_a_container") else vmurl
            pprint(msg)
        else:
            pprint(f"Opening url {vmurl}")
            webbrowser.open(vmurl, new=2, autoraise=True)

    def serialconsole(self, name, web=False):
        return self.console(name)

    def _get_default_network(self, node):
        networks = self.list_networks()
        for n_name, n in networks.items():
            if n_name.split("/")[0] == node:
                return n_name.split("/")[1]

    def update_metadata(self, name, metatype, metavalue, append=False):
        vm = self._get_vm(name)
        if vm is None:
            return {"result": "failure", "reason": f"VM {name} not found."}
        vm_config = vm.config.get()
        description = self._parse_notes(vm_config.get("description"))
        if metatype not in description or metavalue is None or description[metatype] != metavalue:
            if metavalue is None:
                del description[metatype]
            else:
                description[metatype] = metavalue
            description = [f"{key}={description[key]}" for key in description]
            description = ','.join(description)
            vm_data = {'name': name, 'description': description}
            self._wait_for(vm.config.post(**vm_data))

    def _upload_file(self, node_ip, path, data):
        target_dir = "/var/lib/vz/snippets/"

        # Ensure the target directory exists on the remote system
        try:
            check_call(
                f"ssh -q root@{node_ip} 'mkdir -p {target_dir} && test -d {target_dir}'",
                shell=True
            )
        except CalledProcessError as e:
            raise RuntimeError(f"Failed to ensure directory {target_dir} exists on {node_ip}: {e}")

        # Create a temporary file and upload it
        with TemporaryDirectory() as tmpdir:
            temp_file_path = os.path.join(tmpdir, path)
            with open(temp_file_path, "w") as f:
                f.write(data)

            scp_cmd = f"scp -q {temp_file_path} root@{node_ip}:{target_dir}{path}"
            try:
                return call(scp_cmd, shell=True)
            except CalledProcessError as e:
                raise RuntimeError(f"Failed to upload file {temp_file_path} to {target_dir}: {e}")

    def _get_current_disks(self, vm_config):
        disks = []
        for entry in vm_config:
            if re.findall(r"(scsi|ide|virtio)\d+", entry):
                device = entry
                disk_data = vm_config[entry].split(',')
                path = disk_data[0]
                disktype, disksize = disk_data[-1].split('=')
                disksize = 1 if disksize.endswith('M') else int(disksize.replace('G', ''))
                if disktype == 'media':
                    continue
                disks.append({'name': device, 'size': disksize, 'path': path, 'full': vm_config[entry]})
        return disks

    def _get_lxc(self, name):
        for container in self.conn.nodes(self.node).lxc.get():
            if container['name'] == name:
                container_id = container['vmid']
                return self.conn.nodes(self.node).lxc(container_id)

    def create_container(self, name, image, nets=None, cmds=[], ports=[], volumes=[], environment=[], label=None,
                         overrides={}):
        print("NOT IMPLEMENTED")
        return {'result': 'success'}

    def delete_container(self, name):
        container = self._get_lxc(name)
        if container is None:
            return {"result": "failure", "reason": f"Container {name} not found."}
        self._wait_for(container.status.stop.post())
        self._wait_for(container.delete())
        return {'result': 'success'}

    def start_container(self, name):
        container = self._get_lxc(name)
        if container is None:
            return {"result": "failure", "reason": f"Container {name} not found."}
        self._wait_for(container.status.start.post())
        return {'result': 'success'}

    def stop_container(self, name):
        container = self._get_lxc(name)
        if container is None:
            return {"result": "failure", "reason": f"Container {name} not found."}
        self._wait_for(container.status.stop.post())
        return {'result': 'success'}

    def console_container(self, name):
        container = self._get_lxc(name)
        if container is None:
            error(f"Container {name} not found")
            return {'result': 'failure', 'reason': f"Container {name} not found"}
        cont_id = container.status.current.get().get('vmid')
        container_url = f"https://{self.host}:8006/?console=lxc&xtermjs=1&vmid={cont_id}&vmname={name}&node=pve&cmd="
        if self.debug or os.path.exists("/i_am_a_container"):
            msg = f"Open the following url:\n{container_url}" if os.path.exists("/i_am_a_container") else container_url
            pprint(msg)
        else:
            pprint(f"Opening url {container_url}")
            webbrowser.open(container_url, new=2, autoraise=True)
        return {'result': 'success'}

    def list_containers(self):
        containers = []
        for container in self.conn.nodes(self.node).lxc.get():
            name = container['name']
            state = container['status']
            state = 'up' if state.split(' ')[0].startswith('running') else 'down'
            lxc_info = self._get_lxc_info(name, self._get_vm(name).config.get())
            ip, ports, deploy = lxc_info['ip'], '', ''
            plan, image = lxc_info.get('plan', ''), lxc_info.get('image', '')
            containers.append([name, state, image, plan, ip, ports, deploy])
        return containers

    def exists_container(self, name):
        return len([cont for cont in self.conn.nodes(self.node).lxc.get() if cont['name'] == name]) > 0

    def list_images(self):
        images = []
        for storage in self.conn.nodes(self.node).storage.get(content="vztmpl"):
            templates = self.conn.nodes(self.node).storage(storage['storage']).content.get()
            images.extend([item['volid'] for item in templates if item.get("content") == "vztmpl"])
        return sorted(images)

    def _get_lxc_info(self, name, config=None):
        ips = []
        yamlinfo = {'name': config['hostname'], 'memory': config['memory'], 'numcpus': config['cores'], 'nets': []}
        for net in [key for key in config if key.startswith('net')]:
            net = self._parse_notes(config[net])
            device, mac, network, networktype = net['name'], net['hwaddr'].lower(), net['bridge'], net['type']
            yamlinfo['nets'].append({'device': device, 'mac': mac.lower(), 'net': network, 'type': networktype})
            ip = net['ip'].split('/')[0]
            if 'ip' not in yamlinfo:
                yamlinfo['ip'] = ip
            ips.append(ip)
        if len(ips) > 1:
            yamlinfo['ips'] = ips
        path, size = config['rootfs'].split(',')
        disksize = size.replace('size=', '')
        disk = {'device': 'rootfs', 'size': disksize, 'format': 'raw', 'type': 'virtio', 'path': path}
        yamlinfo['disks'] = [disk]
        if 'description' in config:
            description = self._parse_notes(config['description'])
            if 'plan' in description:
                yamlinfo['plan'] = description['plan']
            if 'image' in description:
                yamlinfo['image'] = description['image']
        return yamlinfo

    def detach_disks(self, name):
        print("not implemented")
        return {'result': 'success'}



from base64 import b64encode
from glob import glob
from ipaddress import ip_network
from kvirt.common import success, error, pprint, info2, container_mode, warning, fix_typos, olm_app
from kvirt.common import get_oc, pwd_path, get_installer_rhcos, get_ssh_pub_key, start_baremetal_hosts_with_iso
from kvirt.common import deploy_cloud_storage, patch_ingress_controller_wildcard
from kvirt.cluster.openshift import get_ci_installer, get_downstream_installer, get_installer_version
from kvirt.cluster.openshift import same_release_images, process_apps, offline_image
import json
import os
import re
import socket
import sys
from shutil import which, copy2
from subprocess import call
from tempfile import TemporaryDirectory, NamedTemporaryFile
from time import sleep
from uuid import UUID
from urllib.parse import urlparse
from urllib.request import urlopen, Request
from yaml import safe_dump, safe_load

virt_providers = ['kvm', 'kubevirt', 'ovirt', 'openstack', 'proxmox', 'vsphere']
cloud_providers = ['aws', 'azure', 'gcp', 'ibm', 'hcloud']


def update_hypershift_etc_hosts(cluster, domain, ingress_ip):
    console_entry = f"console-openshift-console.apps.{cluster}.{domain}"
    if not os.path.exists("/i_am_a_container"):
        hosts = open("/etc/hosts").readlines()
        wrongingresses = [e for e in hosts if not e.startswith('#') and console_entry in e and ingress_ip not in e]
        for wrong in wrongingresses:
            warning(f"Cleaning wrong entry {wrong} in /etc/hosts")
            call(f"sudo sed -i '/{wrong.strip()}/d' /etc/hosts", shell=True)
        hosts = open("/etc/hosts").readlines()
        correct = [e for e in hosts if not e.startswith('#') and console_entry in e and ingress_ip in e]
        if not correct:
            entries = [f"{x}.{cluster}.{domain}" for x in ['console-openshift-console.apps',
                                                           'oauth-openshift.apps',
                                                           'prometheus-k8s-openshift-monitoring.apps']]
            entries = ' '.join(entries)
            call(f"sudo sh -c 'echo {ingress_ip} {entries} >> /etc/hosts'", shell=True)
    else:
        entries = [f"{x}.{cluster}.{domain}" for x in ['console-openshift-console.apps',
                                                       'oauth-openshift.apps',
                                                       'prometheus-k8s-openshift-monitoring.apps']]
        entries = ' '.join(entries)
        call(f"sh -c 'echo {ingress_ip} {entries} >> /etc/hosts'", shell=True)
        if os.path.exists('/etcdir/hosts'):
            call(f"sh -c 'echo {ingress_ip} {entries} >> /etcdir/hosts'", shell=True)
        else:
            warning("Make sure to have the following entry in your /etc/hosts")
            warning(f"{ingress_ip} {entries}")


def valid_uuid(uuid):
    try:
        UUID(uuid)
        return True
    except:
        return False


def get_info(url, user, password):
    match = re.match('.*/redfish/v1/Systems/(.*)', url)
    if '/redfish/v1/Systems' in url and\
       (valid_uuid(os.path.basename(url)) or (match is not None and len(match.group(1).split('/')) == 2)):
        model = 'virtual'
        user = user or 'fake'
        password = password or 'fake'
        url = url.replace('http', 'redfish-virtualmedia+http')
        return url, user, password
    oem_url = f"https://{url}" if '://' not in url else url
    p = urlparse(oem_url)
    headers = {'Accept': 'application/json'}
    request = Request(f"https://{p.netloc}/redfish/v1", headers=headers)
    oem = json.loads(urlopen(request).read())['Oem']
    model = "dell" if 'Dell' in oem else "hp" if 'Hpe' in oem else 'supermicro' if 'Supermicro' in oem else 'N/A'
    if '://' not in url:
        if model in ['hp', 'supermicro']:
            url = f"https://{url}/redfish/v1/Systems/1"
        elif model == 'dell':
            url = f"https://{url}/redfish/v1/Systems/System.Embedded.1"
            user = user or 'root'
            password = password or 'calvin'
        else:
            print(f"Invalid url {url}")
            sys.exit(0)
    if model in ['hp', 'hpe', 'supermicro']:
        url = url.replace('https://', 'redfish-virtualmedia')
    elif model == 'dell':
        url = url.replace('https://', 'idrac-virtualmedia')
    return url, user, password


def create_bmh_objects(config, plandir, cluster, namespace, baremetal_hosts, overrides={}):
    clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
    uefi = overrides.get('uefi', True)
    nmstatedata = ''
    with open(f"{clusterdir}/bmcs.yml", 'w') as f:
        for index, host in enumerate(baremetal_hosts):
            bmc_url = host.get('url') or host.get('bmc_url')
            if bmc_url is None:
                warning("Skipping entry {index} in baremetal_hosts array as it misses url")
                continue
            bmc_user = host.get('username') or host.get('user') or host.get('bmc_username') or host.get('bmc_user')\
                or overrides.get('bmc_user') or overrides.get('bmc_username')\
                or overrides.get('user') or overrides.get('username')
            bmc_password = host.get('password') or host.get('bmc_password') or overrides.get('bmc_password')
            bmc_mac = host.get('mac') or host.get('bmc_mac')
            if bmc_mac is None:
                warning("Skipping entry {index} in baremetal_hosts array as it misses mac")
                continue
            bmc_uefi = host.get('uefi') or host.get('bmc_uefi') or uefi
            bmc_name = host.get('name') or host.get('bmc_name') or f'{cluster}-node-{index}'
            bmc_url, bmc_user, bmc_password = get_info(bmc_url, bmc_user, bmc_password)
            bmc_overrides = {'url': bmc_url, 'user': bmc_user, 'password': bmc_password, 'name': bmc_name,
                             'uefi': bmc_uefi, 'namespace': namespace, 'cluster': cluster,
                             'mac': bmc_mac}
            bmc_data = config.process_inputfile(cluster, f"{plandir}/bmc.yml.j2", overrides=bmc_overrides)
            f.write(bmc_data)
            if 'addresses' in host:
                host_data = host.copy()
                host_data['name'] = bmc_name
                host_data['cluster'] = cluster
                host_data['namespace'] = namespace
                if 'dns_resolver' not in host_data:
                    warning("Skipping nmstate for entry {index} as dns_resolver isnt defined")
                    continue
                nmstatedata += config.process_inputfile(cluster, f'{plandir}/nmstateconfig.yml.j2', overrides=host_data)
                nmstatedata += '---\n'
    bmccmd = f"oc create -f {clusterdir}/bmcs.yml"
    call(bmccmd, shell=True)
    if nmstatedata != '':
        with open(f"{clusterdir}/nmstateconfig.yml", 'w') as f:
            f.write(nmstatedata)
        nmcmd = f"oc create -f {clusterdir}/nmstateconfig.yml"
        call(nmcmd, shell=True)


def handle_baremetal_iso(config, plandir, cluster, data, baremetal_hosts=[]):
    baremetal_iso_overrides = data.copy()
    baremetal_iso_overrides['noname'] = True
    baremetal_iso_overrides['workers'] = 1
    result = config.plan(cluster, inputfile=f'{plandir}/kcli_plan.yml', overrides=baremetal_iso_overrides,
                         onlyassets=True)
    iso_data = result['assets'][0]
    with open('iso.ign', 'w') as f:
        f.write(iso_data)
    iso_pool = data.get('pool') or config.pool
    config.create_openshift_iso(cluster, ignitionfile='iso.ign', installer=True)
    if baremetal_hosts:
        iso_pool_path = config.k.get_pool_path(iso_pool)
        chmodcmd = f"chmod 666 {iso_pool_path}/{cluster}-worker.iso"
        call(chmodcmd, shell=True)
        pprint("Creating httpd deployment to host iso for baremetal workers")
        httpdcmd = f"oc create -f {plandir}/httpd.yaml"
        call(httpdcmd, shell=True)
        svcip_cmd = 'oc get node -o yaml'
        svcip = safe_load(os.popen(svcip_cmd).read())['items'][0]['status']['addresses'][0]['address']
        svcport_cmd = 'oc get svc -n default httpd-kcli-svc -o yaml'
        svcport = safe_load(os.popen(svcport_cmd).read())['spec']['ports'][0]['nodePort']
        podname = os.popen('oc -n default get pod -l app=httpd-kcli -o name').read().split('/')[1].strip()
        try:
            call(f"oc wait -n default --for=condition=Ready pod/{podname}", shell=True)
        except Exception as e:
            error(f"Hit {e}")
            sys.exit(1)
        copycmd = f"oc -n default cp {iso_pool_path}/{cluster}-worker.iso {podname}:/var/www/html"
        call(copycmd, shell=True)
        return f'http://{svcip}:{svcport}/{cluster}-worker.iso'


def process_nodepools(config, plandir, namespace, cluster, platform, nodepools, image, overrides, manifests=[]):
    clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
    nodepools = [{'name': entry} if isinstance(entry, str) else entry for entry in nodepools]
    nodepool_data = {'nodepools': nodepools, 'namespace': namespace, 'cluster': cluster,
                     'nodepool_image': image, 'platform': platform, 'workers': overrides.get('workers', 2)}
    kubevirt_data = {'numcpus': overrides.get('numcpus', 8), 'memory': overrides.get('memory', 6144),
                     'disk_size': overrides.get('disk_size', 30)}
    nodepool_data.update(kubevirt_data)
    if manifests:
        nodepool_data['manifests'] = manifests
    nodepoolfile = config.process_inputfile(cluster, f"{plandir}/nodepools.yaml", overrides=nodepool_data)
    with open(f"{clusterdir}/nodepools.yaml", 'w') as f:
        f.write(nodepoolfile)
    cmcmd = f"oc create -f {clusterdir}/nodepools.yaml"
    call(cmcmd, shell=True)
    if platform is not None:
        return
    for entry in nodepools:
        nodepool = entry['name']
        pprint(f"Waiting before ignition data for nodepool {nodepool} is available")
        user_data = f"user-data-{cluster}"
        cmd = f"until oc -n {namespace}-{cluster} get secret | grep {user_data} >/dev/null 2>&1 ; do sleep 1 ; done"
        call(cmd, shell=True)
        ignition_data = {'namespace': namespace, 'cluster': cluster, 'nodepool': nodepool}
        ignitionscript = config.process_inputfile(cluster, f"{plandir}/ignition.sh", overrides=ignition_data)
        with open(f"{clusterdir}/ignition_{nodepool}.sh", 'w') as f:
            f.write(ignitionscript)
        ignition_worker = f"{clusterdir}/nodepool_{nodepool}.ign"
        timeout = 0
        while True:
            if os.path.exists(ignition_worker):
                break
            else:
                sleep(30)
                timeout += 30
                if timeout > 300:
                    msg = "Timeout trying to retrieve worker ignition"
                    return {'result': 'failure', 'reason': msg}
            call(f'bash {clusterdir}/ignition_{nodepool}.sh', shell=True)


def scale(config, plandir, cluster, overrides):
    storedparameters = overrides.get('storedparameters', True)
    provider = config.type
    plan = cluster
    data = {'cluster': cluster, 'kube': cluster, 'kubetype': 'hypershift', 'namespace': 'clusters',
            'disk_size': 30, 'numcpus': 8, 'memory': 16384}
    data['basedir'] = '/workdir' if container_mode() else '.'
    cluster = data['cluster']
    clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
    if not os.path.exists(clusterdir):
        warning(f"Creating {clusterdir} from your input (auth creds will be missing)")
        data['client'] = config.client
        overrides['cluster'] = cluster
        overrides['clusterdir'] = clusterdir
        plan = overrides.get('plan') or plan
        namespace = data['namespace']
        assisted = cluster in os.popen(f'oc get -n {namespace}-{cluster} infraenv {cluster}').read().strip()
        if not assisted:
            if 'ingress_ip' not in overrides and provider != 'kubevirt':
                msg = "Missing ingress_ip..."
                return {'result': 'failure', 'reason': msg}
            domain = overrides.get('domain')
            if domain is None:
                msg = "Missing domain..."
                return {'result': 'failure', 'reason': msg}
            if 'management_ingress_domain' not in overrides:
                overrides['management_ingress_domain'] = f'apps.{cluster}.{domain}'
            ignitionscript = config.process_inputfile(cluster, f"{plandir}/ignition.sh", overrides=overrides)
            with open(f"{clusterdir}/ignition.sh", 'w') as f:
                f.write(ignitionscript)
            call(f'bash {clusterdir}/ignition.sh', shell=True)
        else:
            os.mkdir(clusterdir)
            data['assisted'] = True
    old_extra_nodepools = []
    if storedparameters and os.path.exists(f"{clusterdir}/kcli_parameters.yml"):
        with open(f"{clusterdir}/kcli_parameters.yml", 'r') as install:
            installparam = safe_load(install)
            data.update(installparam)
            plan = installparam.get('plan', plan)
            old_extra_nodepools = installparam.get('extra_nodepools', old_extra_nodepools)
    data.update(overrides)
    namespace = data['namespace']
    nodepools = [p for p in data.get('extra_nodepools') if p not in old_extra_nodepools]
    if nodepools:
        platform = data.get('platform')
        nodepool_cmd = "oc get nodepool -n %s %s -o jsonpath='{.spec.release.image}'" % (namespace, cluster)
        nodepool_image = os.popen(nodepool_cmd).read()
        process_nodepools(config, plandir, namespace, cluster, platform, nodepools, nodepool_image, overrides)
    nodepool = overrides.get('nodepool') or cluster
    if os.popen(f"oc get nodepool -n {namespace} {nodepool} -o name").read() == "":
        msg = f"Issue with nodepool {nodepool}..."
        return {'result': 'failure', 'reason': msg}
    if os.path.exists(f"{clusterdir}/kubeconfig.mgmt"):
        os.environ['KUBECONFIG'] = f"{clusterdir}/kubeconfig.mgmt"
    if 'KUBECONFIG' not in os.environ:
        warning("KUBECONFIG not set...Using .kube/config instead")
    platform = data.get('platform')
    assisted = platform == 'assisted'
    kubevirt = platform == 'kubevirt'
    with open(f"{clusterdir}/kcli_parameters.yml", 'w') as paramfile:
        safe_dump(data, paramfile, default_flow_style=False, encoding='utf-8', allow_unicode=True)
    pprint(f"Scaling on client {config.client}")
    worker_overrides = data.copy()
    workers = worker_overrides.get('workers', 2)
    if kubevirt:
        cmcmd = f"oc -n {namespace}-{cluster} scale nodepool {nodepool} --replicas {workers}"
        call(cmcmd, shell=True)
        return {'result': 'success'}
    os.chdir(os.path.expanduser("~/.kcli"))
    old_baremetal_hosts = installparam.get('baremetal_hosts', [])
    new_baremetal_hosts = overrides.get('baremetal_hosts', [])
    if assisted:
        ksushy_ip = data['assisted_vms_ksushy_ip']
        ksushy_url = f'http://{ksushy_ip}:9000/redfish/v1/Systems/{config.client}'
        old_physical_baremetal_hosts = [h for h in old_baremetal_hosts if ksushy_ip not in h['url']]
        assisted_vms_number = workers - len(old_physical_baremetal_hosts + new_baremetal_hosts)
        if assisted_vms_number > 0:
            data['assisted_vms_number'] = assisted_vms_number
            data['disk_size'] = max(data['disk_size'], 200)
            data['numcpus'] = max(data['numcpus'], 16)
            data['memory'] = max(data['memory'], 20480)
            worker_threaded = data.get('threaded', False) or data.get('assisted_vms_threaded', False)
            result = config.plan(plan, inputfile=f'{plandir}/kcli_plan_assisted.yml', overrides=data,
                                 threaded=worker_threaded)
            if result['result'] != 'success':
                return result
            vms = result['newvms']
            virtual_hosts = []
            for vm in vms:
                new_mac = config.k.info(vm)['nets'][0]['mac']
                virtual_hosts.append({'url': f'{ksushy_url}/{vm}', 'mac': new_mac})
            new_baremetal_hosts.extend(virtual_hosts)
    baremetal_hosts = [entry for entry in new_baremetal_hosts if entry not in old_baremetal_hosts]
    if baremetal_hosts:
        if assisted:
            all_baremetal_hosts = old_baremetal_hosts + baremetal_hosts
            create_bmh_objects(config, plandir, cluster, namespace, all_baremetal_hosts, overrides)
            replicas_cmd = "oc get nodepool -n %s %s -o jsonpath='{.spec.replicas}'" % (namespace, nodepool)
            replicas = int(os.popen(replicas_cmd).read()) + len(baremetal_hosts)
            pprint(f"Setting replicas to {replicas} in nodepool {nodepool}")
            cmcmd = f"oc -n {data['namespace']} scale nodepool {nodepool} --replicas {replicas}"
            call(cmcmd, shell=True)
            return {'result': 'success'}
        else:
            if not old_baremetal_hosts:
                iso_url = handle_baremetal_iso(config, plandir, cluster, data, baremetal_hosts)
            else:
                svcip_cmd = 'oc get node -o yaml'
                svcip = safe_load(os.popen(svcip_cmd).read())['items'][0]['status']['addresses'][0]['address']
                svcport_cmd = 'oc get svc -n default httpd-kcli-svc -o yaml'
                svcport = safe_load(os.popen(svcport_cmd).read())['spec']['ports'][0]['nodePort']
                iso_url = f'http://{svcip}:{svcport}/{cluster}-worker.iso'
            result = start_baremetal_hosts_with_iso(baremetal_hosts, iso_url, overrides=overrides, debug=config.debug)
            if result['result'] != 'success':
                return result
            worker_overrides['workers'] = workers - len(new_baremetal_hosts)
    if platform is not None:
        return {'result': 'success'}
    if worker_overrides.get('workers', 2) <= 0:
        return {'result': 'success'}
    threaded = data.get('threaded', False) or data.get('workers_threaded', False)
    return config.plan(plan, inputfile=f'{plandir}/kcli_plan.yml', overrides=worker_overrides, threaded=threaded)


def create(config, plandir, cluster, overrides):
    log_level = 'debug' if config.debug else 'info'
    k = config.k
    provider = config.type
    arch = k.get_capabilities()['arch'] if provider == 'kvm' else 'x86_64'
    arch_tag = 'arm64' if arch in ['aarch64', 'arm64'] else 'latest'
    overrides['arch_tag'] = arch_tag
    if 'KUBECONFIG' not in os.environ:
        warning("KUBECONFIG not set...Using .kube/config instead")
    elif not os.path.isabs(os.environ['KUBECONFIG']):
        os.environ['KUBECONFIG'] = f"{os.getcwd()}/{os.environ['KUBECONFIG']}"
        if not os.path.exists(os.environ['KUBECONFIG']):
            return {'result': 'failure', 'reason': "Kubeconfig not found. Leaving..."}
    data = safe_load(open(f'{plandir}/kcli_default.yml'))
    data.update(overrides)
    fix_typos(data)
    retries = data.get('retries')
    data['cluster'] = cluster
    data['kube'] = data['cluster']
    data['kubetype'] = 'hypershift'
    ignore_hosts = data['ignore_hosts'] or data['sslip']
    pprint(f"Deploying cluster {cluster}")
    plan = cluster
    upstream = data['upstream']
    platform = data.get('platform')
    assisted = platform == 'assisted'
    data['assisted'] = assisted
    kubevirt = platform == 'kubevirt'
    data['kubevirt'] = kubevirt
    if platform not in [None, 'kubevirt', 'assisted']:
        return {'result': 'failure', 'reason': "Incorrect platform. Choose between None, kubevirt and assisted"}
    baremetal_iso = data.get('baremetal_iso', False)
    baremetal_hosts = data.get('baremetal_hosts')
    async_install = data.get('async')
    notify = data.get('notify')
    autoscale = data.get('autoscale')
    sslip = data.get('sslip')
    domain = data.get('domain')
    original_domain = domain
    data['original_domain'] = domain
    apps = overrides.get('apps', [])
    workers = data.get('workers')
    use_management_version = False
    version = data.get('version')
    if version not in ['candidate', 'stable', 'ci', 'nightly', 'latest', 'cluster']:
        return {'result': 'failure', 'reason': f"Invalid version {version}"}
    elif version == 'cluster':
        use_management_version = True
        version = 'stable'
    coredns = data.get('coredns')
    tag = data.get('tag')
    if str(tag) == '4.1':
        tag = '4.10'
        data['tag'] = tag
    pull_secret = os.path.expanduser(pwd_path(data.get('pull_secret')))
    if not os.path.exists(pull_secret):
        return {'result': 'failure', 'reason': f"Missing pull secret file {pull_secret}"}
    network_type = data['network_type']
    if network_type not in ['Calico', 'OVNKubernetes']:
        data['network_type'] = 'Other'
    cluster = data.get('cluster')
    namespace = data.get('namespace')
    clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
    if os.path.exists(clusterdir):
        return {'result': 'failure', 'reason': f"Remove existing directory {clusterdir} or use --force"}
    else:
        os.makedirs(f"{clusterdir}/auth")
    if which('oc') is None:
        get_oc()
    default_sc = False
    for sc in safe_load(os.popen('oc get sc -o yaml').read()).get('items', []):
        if 'annotations' in sc['metadata']\
           and 'storageclass.kubernetes.io/is-default-class' in sc['metadata']['annotations']\
           and sc['metadata']['annotations']['storageclass.kubernetes.io/is-default-class'] == 'true':
            pprint(f"Using default class {sc['metadata']['name']}")
            default_sc = True
    if not default_sc:
        return {'result': 'failure', 'reason': "Default Storage class not found. Leaving..."}
    kubevirt_crd_cmd = 'oc get crd hyperconvergeds.hco.kubevirt.io -o yaml 2>/dev/null'
    if kubevirt and safe_load(os.popen(kubevirt_crd_cmd).read()) is None:
        warning("Kubevirt not fully installed. Installing it for you")
        app_name, source, channel, csv, description, x_namespace, channels, crds = olm_app('kubevirt-hyperconverged')
        app_data = {'name': app_name, 'source': source, 'channel': channel, 'namespace': x_namespace, 'csv': csv}
        result = config.create_app_openshift(app_name, app_data)
        if result != 0:
            return {'result': 'failure', 'reason': "Couldnt install kubevirt properly"}
    kubeconfig = os.environ.get('KUBECONFIG')
    kubeconfigdir = os.path.dirname(kubeconfig) if kubeconfig is not None else os.path.expanduser("~/.kube")
    kubeconfig = os.path.basename(kubeconfig) if kubeconfig is not None else 'config'
    hosted_crd_cmd = 'oc get crd hostedclusters.hypershift.openshift.io -o yaml 2>/dev/null'
    assisted_crd_cmd = 'oc -n multicluster-engine get pod -l app=assisted-service -o name 2>/dev/null'
    need_hypershift = safe_load(os.popen(hosted_crd_cmd).read()) is None
    need_assisted = assisted and safe_load(os.popen(assisted_crd_cmd).read()) is None
    need_mce = (need_hypershift and not upstream) or need_assisted
    if need_mce:
        mce_hypershift = need_hypershift and not upstream
        if mce_hypershift:
            warning("Hypershift needed. Installing it for you")
        if need_assisted:
            warning("Assisted needed. Installing it for you")
        app_name, source, channel, csv, description, x_namespace, channels, crds = olm_app('multicluster-engine')
        app_data = {'name': app_name, 'source': source, 'channel': channel, 'namespace': x_namespace, 'csv': csv,
                    'mce_hypershift': mce_hypershift, 'assisted': need_assisted, 'version': version, 'tag': tag,
                    'pull_secret': pull_secret}
        result = config.create_app_openshift(app_name, app_data)
        if result != 0:
            return {'result': 'failure', 'reason': "Couldnt install mce properly"}
        sleep(240)
        if assisted and safe_load(os.popen(assisted_crd_cmd).read()) is None:
            return {'result': 'failure', 'reason': "Couldnt install assisted via mce properly"}
    if need_hypershift and upstream:
        warning("Hypershift needed. Installing it for you")
        if which('podman') is None:
            return {'result': 'failure', 'reason': "Please install podman first in order to install hypershift"}
        else:
            hypercmd = f"podman pull {data['operator_image']}"
            call(hypercmd, shell=True)
            hypercmd = "podman run -it --rm --security-opt label=disable --entrypoint=/usr/bin/hypershift "
            hypercmd += f"-e KUBECONFIG=/k/{kubeconfig} -v {kubeconfigdir}:/k {data['operator_image']} install"
            hypercmd += f" --hypershift-image {data['operator_image']}"
            call(hypercmd, shell=True)
            sleep(120)
    if safe_load(os.popen(hosted_crd_cmd).read()) is None:
        return {'result': 'failure', 'reason': "Couldnt install hypershift properly"}
    registry = 'quay.io'
    management_image = os.popen("oc get clusterversion version -o jsonpath='{.status.desired.image}'").read()
    prefixes = ['quay.io', 'registry.ci']
    fake_disconnected = overrides.get('fake_disconnected', False)
    if not fake_disconnected and not any(management_image.startswith(p) for p in prefixes):
        disconnected_url = management_image.split('/')[0]
        data['registry'] = disconnected_url
        if disconnected_url not in open(pull_secret).read():
            msg = f"entry for {disconnected_url} missing in pull secret"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        hypershift_info = f"oc adm release info --insecure --pullspecs -a {pull_secret} | "
        hypershift_info += "grep hypershift | awk '{print $2}' | cut -d@ -f2"
        hypershift_tag = os.popen(hypershift_info).read().strip()
        mco_info = f"oc adm release info --insecure --pullspecs -a {pull_secret} | "
        mco_info += "grep machine-config-operator | awk '{print $2'} | cut -d@ -f2"
        mco_tag = os.popen(mco_info).read().strip()
        disconnected_data = {'disconnected_url': disconnected_url, 'hypershift_tag': hypershift_tag, 'mco_tag': mco_tag}
        disconnectedfile = config.process_inputfile(cluster, f"{plandir}/disconnected.sh", overrides=disconnected_data)
        with open(f"{clusterdir}/disconnected.sh", 'w') as f:
            f.write(disconnectedfile)
        call(f'bash {clusterdir}/disconnected.sh', shell=True)
        os.environ['OPENSHIFT_INSTALL_RELEASE_IMAGE_OVERRIDE'] = management_image
        cacmd = "oc -n openshift-config get cm user-ca-bundle -o jsonpath='{.data.ca-bundle\\.crt}'"
        data['ca'] = os.popen(cacmd).read().strip()
        data['operator_disconnected_image'] = f'{disconnected_url}/openshift/release@{hypershift_tag}'
    data['registry'] = registry
    data['basedir'] = '/workdir' if container_mode() else '.'
    supported_data = safe_load(os.popen("oc get cm/supported-versions -o yaml -n hypershift").read())
    if use_management_version:
        tag = os.popen("oc get clusterversion version -o jsonpath='{.status.desired.version}'").read()
        pprint(f"Using tag {tag}")
        if 'ec' in tag:
            version = 'candidate'
    elif supported_data is not None:
        supported_data = supported_data['data']
        supported_versions = supported_versions = supported_data['supported-versions']
        versions = safe_load(supported_versions)['versions']
        if version == 'latest':
            tag = versions[0]
        elif str(tag) not in versions:
            return {'result': 'failure', 'reason': f"Invalid tag {tag}. Choose between {','.join(versions)}"}
    else:
        warning(f"Couldnt verify whether {tag} is a valid tag")
    management_cmd = "oc get ingresscontroller -n openshift-ingress-operator default -o jsonpath='{.status.domain}'"
    management_ingress_domain = os.popen(management_cmd).read()
    data['management_ingress_domain'] = management_ingress_domain
    management_ingress_ip = data.get('management_ingress_ip')
    if management_ingress_ip is None:
        try:
            management_ingress_ip = socket.getaddrinfo('xxx.' + management_ingress_domain, 80,
                                                       proto=socket.IPPROTO_TCP)[0][4][0]
            data['management_ingress_ip'] = management_ingress_ip
        except:
            warning("Couldn't figure out management ingress ip. Using node port instead")
            data['nodeport'] = True
    management_api_ip = data.get('management_api_ip')
    if management_api_ip is None:
        management_ips_cmd = "oc get nodes --no-headers -o jsonpath=\"{.items[*]['status.addresses'][0]['address']}\""
        management_ips = os.popen(management_ips_cmd).read().split(' ')
        if len(management_ips) == 1:
            management_api_ip = management_ips[0]
        else:
            management_api_url = os.popen("oc whoami --show-server").read()
            management_api_domain = urlparse(management_api_url).hostname
            management_api_ip = socket.getaddrinfo(management_api_domain, 6443, proto=socket.IPPROTO_TCP)[0][4][0]
        data['management_api_ip'] = management_api_ip
    pprint(f"Using {management_api_ip} as management api ip")
    pub_key = data.get('pub_key')
    data['pull_secret'] = re.sub(r"\s", "", open(pull_secret).read())
    pub_key = data.get('pub_key') or get_ssh_pub_key()
    keys = data.get('keys', [])
    if pub_key is None:
        if keys:
            warning("Using first key from your keys array")
            pub_key = keys[0]
        else:
            msg = "No usable public key found, which is required for the deployment. Create one using ssh-keygen"
            return {'result': 'failure', 'reason': msg}
    pub_key = os.path.expanduser(pub_key)
    if pub_key.startswith('ssh-'):
        data['pub_key'] = pub_key
    elif os.path.exists(pub_key):
        data['pub_key'] = open(pub_key).read().strip()
    else:
        return {'result': 'failure', 'reason': f"Public key file {pub_key} not found"}
    network = data.get('network')
    ingress_ip = data.get('ingress_ip')
    virtual_router_id = None
    kubevirt_ingress_svc = False
    if ingress_ip is None:
        networkinfo = k.info_network(network)
        if provider == 'kubevirt' or kubevirt:
            domain = management_ingress_domain
            data['domain'] = domain
            pprint(f"Setting domain to {domain}")
            try:
                patch_ingress_controller_wildcard()
            except:
                warning("Couldnt patch ingresscontroller to support wildcards. Assuming it's configured properly")
            if not kubevirt:
                selector = {'kcli/plan': plan, 'kcli/role': 'worker'}
                service_type = "LoadBalancer" if k.access_mode == 'LoadBalancer' else 'NodePort'
                ingress_ip = k.create_service(f"{cluster}-ingress", k.namespace, selector, _type=service_type,
                                              ports=[80, 443])
                kubevirt_ingress_svc = True
                if service_type == 'NodePort':
                    hostname = f"http.apps.{cluster}.{management_ingress_domain}"
                    route_cmd = f"oc -n {k.namespace} create route passthrough --service={cluster}-ingress "
                    route_cmd += f"--hostname={hostname} --wildcard-policy=Subdomain --port=443"
                    call(route_cmd, shell=True)
                elif ingress_ip is None:
                    return {'result': 'failure', 'reason': f"Couldnt gather an ingress_ip from network {network}"}
        elif provider == 'kvm' and networkinfo['type'] == 'routed':
            cidr = networkinfo['cidr']
            ingress_index = 3 if ':' in cidr else -4
            ingress_ip = str(ip_network(cidr)[ingress_index])
            warning(f"Using {ingress_ip} as ingress_ip")
            data['ingress_ip'] = ingress_ip
        else:
            return {'result': 'failure', 'reason': "You need to define ingress_ip in your parameters file"}
    elif kubevirt:
        warning(f"Note that ingress_ip {ingress_ip} won't be configured as you're using kubevirt provider")
    if ingress_ip is not None and data.get('virtual_router_id') is None and not kubevirt_ingress_svc:
        virtual_router_id = hash(cluster) % 254 + 1
        data['virtual_router_id'] = virtual_router_id
        pprint(f"Using keepalived virtual_router_id {virtual_router_id}")
    if ingress_ip is not None and sslip and provider in virt_providers:
        domain = f"{ingress_ip.replace('.', '-').replace(':', '-')}.sslip.io"
        data['domain'] = domain
        pprint(f"Setting domain to {domain}")
    assetsdata = data.copy()
    copy2(f'{kubeconfigdir}/{kubeconfig}', f"{clusterdir}/kubeconfig.mgmt")
    cidr = data.get('cidr', '192.168.122.0/24')
    assetsdata['cidr'] = cidr
    ipv6 = ':' in cidr
    data['ipv6'] = ipv6
    pprint("Creating control plane assets")
    cmcmd = f"oc create ns {namespace} -o yaml --dry-run=client | oc apply -f -"
    call(cmcmd, shell=True)
    icsps = safe_load(os.popen('oc get imagecontentsourcepolicies -o yaml').read())['items']
    if not fake_disconnected and icsps:
        imagecontentsources = []
        for icsp in icsps:
            for mirror_spec in icsp['spec']['repositoryDigestMirrors']:
                source, mirror = mirror_spec['source'], mirror_spec['mirrors'][0]
                imagecontentsources.append({'source': source, 'mirror': mirror})
        assetsdata['imagecontentsources'] = imagecontentsources
    manifests = []
    manifestsdir = pwd_path("manifests")
    if os.path.exists(manifestsdir) and os.path.isdir(manifestsdir):
        for f in glob(f"{manifestsdir}/*.y*ml"):
            mc_name = os.path.basename(f).replace('.yaml', '').replace('.yml', '')
            mc_data = safe_load(open(f))
            if mc_data.get('kind', 'xx') == 'MachineConfig':
                pprint(f"Injecting manifest {f}")
                mc_data = json.dumps(mc_data)
                manifests.append({'name': mc_name, 'data': mc_data})
    ingress_metallb = data['assisted_ingress_metallb']
    if assisted:
        keepalived_yml_data = config.process_inputfile(cluster, f"{plandir}/staticpods/keepalived.yml", overrides=data)
        keepalived_yml_data = b64encode(keepalived_yml_data.encode()).decode("UTF-8")
        keepalived_conf_data = config.process_inputfile(cluster, f"{plandir}/keepalived.conf", overrides=data)
        keepalived_conf_data = b64encode(keepalived_conf_data.encode()).decode("UTF-8")
        assisted_data = {'keepalived_yml_data': keepalived_yml_data, 'keepalived_conf_data': keepalived_conf_data}
        if not sslip and coredns:
            coredns_yml_data = config.process_inputfile(cluster, f"{plandir}/staticpods/coredns.yml", overrides=data)
            coredns_yml_data = b64encode(coredns_yml_data.encode()).decode("UTF-8")
            assisted_data['coredns_yml_data'] = coredns_yml_data
            coredns_conf_data = config.process_inputfile(cluster, f"{plandir}/Corefile", overrides=data)
            coredns_conf_data = b64encode(coredns_conf_data.encode()).decode("UTF-8")
            assisted_data['coredns_conf_data'] = coredns_conf_data
            force_dns_data = config.process_inputfile(cluster, f"{plandir}/99-kcli-forcedns", overrides=data)
            force_dns_data = b64encode(force_dns_data.encode()).decode("UTF-8")
            assisted_data['force_dns_data'] = force_dns_data
        assisted_data['cluster'] = cluster
        assisted_data['sslip'] = sslip
        assisted_data['coredns'] = coredns
        if not ingress_metallb:
            assisted_data = config.process_inputfile(cluster, f'{plandir}/assisted_ingress.yml',
                                                     overrides=assisted_data)
            assisted_data = json.dumps(assisted_data)
            manifests.append({'name': f'assisted-ingress-{cluster}', 'data': assisted_data})
    async_files = []
    async_tempdir = TemporaryDirectory()
    asyncdir = async_tempdir.name
    if notify:
        notifycmd = "cat /shared/results.txt"
        notifycmds, mailcontent = config.handle_notifications(cluster, notifymethods=config.notifymethods,
                                                              pushbullettoken=config.pushbullettoken,
                                                              notifycmd=notifycmd, slackchannel=config.slackchannel,
                                                              slacktoken=config.slacktoken,
                                                              mailserver=config.mailserver,
                                                              mailfrom=config.mailfrom, mailto=config.mailto,
                                                              cluster=True)
        notifyfile = f"{plandir}/99-notifications.yaml"
        notifyfile = config.process_inputfile(cluster, notifyfile, overrides={'registry': registry,
                                                                              'cluster': cluster,
                                                                              'domain': original_domain,
                                                                              'cmds': notifycmds,
                                                                              'mailcontent': mailcontent})
        with open(f"{asyncdir}/99-notifications.yaml", 'w') as f:
            f.write(notifyfile)
        async_files.append({'path': '/etc/kubernetes/99-notifications.yaml',
                            'origin': f"{asyncdir}/99-notifications.yaml"})
    if async_install:
        if apps:
            autolabeller = False
            final_apps = []
            for a in apps:
                if isinstance(a, str) and a == 'users' or (isinstance(a, dict) and a.get('name', '') == 'users'):
                    continue
                elif isinstance(a, str) and a == 'autolabeller'\
                        or (isinstance(a, dict) and a.get('name', '') == 'autolabeller'):
                    autolabeller = True
                elif isinstance(a, str) and a != 'nfs':
                    final_apps.append(a)
                elif isinstance(a, dict) and 'name' in a:
                    final_apps.append(a['name'])
                else:
                    error(f"Invalid app {a}. Skipping")
            appsfile = f"{plandir}/99-apps.yaml"
            apps_data = {'registry': registry, 'apps': final_apps, overrides: safe_dump(overrides)}
            appsfile = config.process_inputfile(cluster, appsfile, overrides=apps_data)
            with open(f"{asyncdir}/99-apps.yaml", 'w') as f:
                f.write(appsfile)
            async_files.append({'path': '/etc/kubernetes/99-apps.yaml',
                                'origin': f"{asyncdir}/99-apps.yaml"})
            appdir = f"{plandir}/apps"
            apps_namespace = {'advanced-cluster-management': 'open-cluster-management',
                              'multicluster-engine': 'multicluster-engine', 'kubevirt-hyperconverged': 'openshift-cnv',
                              'local-storage-operator': 'openshift-local-storage',
                              'ocs-operator': 'openshift-storage', 'odf-lvm-operator': 'openshift-storage',
                              'odf-operator': 'openshift-storage', 'metallb-operator': 'openshift-operators',
                              'autolabeller': 'autorules'}
            if autolabeller:
                final_apps.append('autolabeller')
            for appname in final_apps:
                app_data = data.copy()
                if data.get('apps_install_cr') and os.path.exists(f"{appdir}/{appname}/cr.yml"):
                    app_data['namespace'] = apps_namespace[appname]
                    if original_domain is not None:
                        app_data['domain'] = original_domain
                    cr_content = config.process_inputfile(cluster, f"{appdir}/{appname}/cr.yml", overrides=app_data)
                    rendered = config.process_inputfile(cluster, f"{plandir}/99-apps-cr.yaml",
                                                        overrides={'registry': registry,
                                                                   'app': appname,
                                                                   'cr_content': cr_content})
                    with open(f"{asyncdir}/99-apps-cr.yaml", 'w') as f:
                        f.write(rendered)
                    async_files.append({'path': '/etc/kubernetes/99-app-cr.yaml',
                                        'origin': f"{asyncdir}/99-apps-cr.yaml"})
        if autoscale:
            config.import_in_kube(network=network, dest=f"{clusterdir}", secure=True)
            for entry in ["99-kcli-conf-cm.yaml", "99-kcli-ssh-cm.yaml"]:
                async_files.append({'path': f'/etc/kubernetes/{entry}', 'origin': f"{clusterdir}/{entry}"})
            commondir = os.path.dirname(pprint.__code__.co_filename)
            autoscale_overrides = {'cluster': cluster, 'kubetype': 'hypershift', 'workers': workers, 'replicas': 1,
                                   'sa': 'default'}
            autoscale_data = config.process_inputfile(cluster, f"{commondir}/autoscale.yaml.j2",
                                                      overrides=autoscale_overrides)
            with open(f"{asyncdir}/autoscale.yaml", 'w') as f:
                f.write(autoscale_data)
            async_files.append({'path': '/etc/kubernetes/autoscale.yaml', 'origin': f"{asyncdir}/autoscale.yaml"})
    data['async_files'] = async_files
    if assisted:
        assistedfile = config.process_inputfile(cluster, f"{plandir}/assisted_infra.yml", overrides=assetsdata)
        with open(f"{clusterdir}/assisted_infra.yml", 'w') as f:
            f.write(assistedfile)
        cmcmd = f"oc create -f {clusterdir}/assisted_infra.yml"
        call(cmcmd, shell=True)
    if 'OPENSHIFT_INSTALL_RELEASE_IMAGE_OVERRIDE' in os.environ:
        assetsdata['hosted_image'] = os.environ['OPENSHIFT_INSTALL_RELEASE_IMAGE_OVERRIDE']
    else:
        if use_management_version:
            assetsdata['hosted_image'] = management_image
        elif version in ['candidate', 'stable', 'ci', 'nightly', 'latest']:
            assetsdata['hosted_image'] = offline_image(version=version, tag=tag, pull_secret=pull_secret)
        else:
            return {'result': 'failure', 'reason': f"Invalid version {version}"}
    hostedclusterfile = config.process_inputfile(cluster, f"{plandir}/hostedcluster.yaml", overrides=assetsdata)
    with open(f"{clusterdir}/hostedcluster.yaml", 'w') as f:
        f.write(hostedclusterfile)
    cmcmd = f"oc create -f {clusterdir}/hostedcluster.yaml"
    call(cmcmd, shell=True)
    which_openshift = which('openshift-install')
    openshift_dir = os.path.dirname(which_openshift) if which_openshift is not None else '.'
    if not same_release_images(version=version, tag=tag, pull_secret=pull_secret, path=openshift_dir):
        if version in ['ci', 'nightly']:
            nightly = version == 'nigthly'
            run = get_ci_installer(pull_secret, tag=tag, nightly=nightly)
        elif version in ['candidate', 'stable', 'latest']:
            run = get_downstream_installer(version=version, tag=tag, pull_secret=pull_secret)
        else:
            return {'result': 'failure', 'reason': f"Invalid version {version}"}
        if run != 0:
            msg = "Couldn't download openshift-install"
            return {'result': 'failure', 'reason': msg}
        pprint("Move downloaded openshift-install somewhere in your PATH if you want to reuse it")
    elif which_openshift is not None:
        pprint("Using existing openshift-install found in your PATH")
    else:
        pprint("Reusing matching openshift-install")
    os.environ["PATH"] = f'{os.getcwd()}:{os.environ["PATH"]}'
    if 'OPENSHIFT_INSTALL_RELEASE_IMAGE_OVERRIDE' in os.environ:
        nodepool_image = os.environ['OPENSHIFT_INSTALL_RELEASE_IMAGE_OVERRIDE']
    elif use_management_version:
        nodepool_image = management_image
    else:
        INSTALLER_VERSION = get_installer_version()
        pprint(f"Using installer version {INSTALLER_VERSION}")
        nodepool_image = os.popen("openshift-install version | grep 'release image' | cut -f3 -d' '").read().strip()
    if assisted:
        data['disk_size'] = max(data['disk_size'], 200)
        data['numcpus'] = max(data['numcpus'], 16)
        data['memory'] = max(data['memory'], 20480)
        worker_threaded = data.get('threaded', False)
        assisted_vms_number = workers - len(baremetal_hosts)
        if assisted_vms_number > 0:
            data['assisted_vms_number'] = assisted_vms_number
            result = config.plan(plan, inputfile=f'{plandir}/kcli_plan_assisted.yml', overrides=data,
                                 threaded=worker_threaded)
            if result['result'] != 'success':
                return result
            vms = result['newvms']
            ksushy_ip = data['assisted_vms_ksushy_ip']
            ksushy_url = f'http://{ksushy_ip}:9000/redfish/v1/Systems/{config.client}'
            virtual_hosts = []
            for vm in vms:
                new_mac = config.k.info(vm)['nets'][0]['mac']
                virtual_hosts.append({'url': f'{ksushy_url}/{vm}', 'mac': new_mac})
            baremetal_hosts.extend(virtual_hosts)
        create_bmh_objects(config, plandir, cluster, namespace, baremetal_hosts, overrides)
        agents = len(baremetal_hosts) or 2
        pprint(f"Waiting for {agents} agents to appear")
        agent_ns = f"{namespace}-{cluster}"
        call(f'until [ "$(oc -n {agent_ns} get agent -o name | wc -l | xargs)" -eq "{agents}" ] ; do sleep 1 ; done',
             shell=True)
        pprint("Waiting 2mn to avoid race conditions when creating nodepool")
        sleep(120)
    elif not kubevirt:
        image = data.get('image')
        if image is None:
            image_type = 'openstack' if data.get('kvm_openstack', True) and provider == 'kvm' else provider
            region = config.k.region if provider == 'aws' else None
            image_url = get_installer_rhcos(_type=image_type, region=region, arch=arch)
            if provider in ['aws', 'gcp']:
                image = image_url
            else:
                image = os.path.basename(os.path.splitext(image_url)[0])
                if provider == 'ibm':
                    image = image.replace('.', '-').replace('_', '-').lower()
                images = [v for v in k.volumes() if image in v]
                if not images:
                    result = config.download_image(pool=config.pool, image=image, url=image_url,
                                                   size=data.get('kubevirt_disk_size'))
                    if result['result'] != 'success':
                        return result
            pprint(f"Using image {image}")
            data['image'] = image
        elif provider == 'kubevirt' and '/' in image:
            warning(f"Assuming image {image} is available")
        else:
            pprint(f"Checking if image {image} is available")
            images = [v for v in k.volumes() if image in v]
            if not images:
                msg = f"Missing {image}. Indicate correct image in your parameters file..."
                return {'result': 'failure', 'reason': msg}
    extra_nodepools = data.get('extra_nodepools', [])
    nodepools = [cluster] + extra_nodepools
    process_nodepools(config, plandir, namespace, cluster, platform, nodepools, nodepool_image, overrides, manifests)
    if 'name' in data:
        del data['name']
    with open(f"{clusterdir}/kcli_parameters.yml", 'w') as p:
        installparam = overrides.copy()
        installparam['client'] = config.client
        installparam['plan'] = plan
        installparam['cluster'] = cluster
        installparam['kubetype'] = 'hypershift'
        installparam['management_api_ip'] = management_api_ip
        installparam['management_ingress_domain'] = management_ingress_domain
        if management_ingress_ip is not None:
            installparam['management_ingress_ip'] = management_ingress_ip
        if ingress_ip is not None:
            installparam['ingress_ip'] = ingress_ip
        if virtual_router_id is not None:
            installparam['virtual_router_id'] = virtual_router_id
        installparam['platform'] = platform
        if platform is None:
            installparam['image'] = image
        installparam['ipv6'] = ipv6
        installparam['original_domain'] = data['original_domain']
        if baremetal_hosts:
            installparam['baremetal_hosts'] = baremetal_hosts
        installparam['registry'] = data.get('registry', 'quay.io')
        installparam['extra_nodepools'] = extra_nodepools
        safe_dump(installparam, p, default_flow_style=False, encoding='utf-8', allow_unicode=True)
    autoapproverpath = f'{clusterdir}/autoapprovercron.yml'
    autoapprover = config.process_inputfile(cluster, f"{plandir}/autoapprovercron.yml", overrides=data)
    with open(autoapproverpath, 'w') as f:
        f.write(autoapprover)
    call(f"oc apply -f {autoapproverpath}", shell=True)
    if platform is None:
        if provider in cloud_providers + ['openstack']:
            copy2(f"{clusterdir}/nodepool_{cluster}.ign", f"{clusterdir}/nodepool_{cluster}.ign.ori")
            bucket = f"{cluster}-{domain.replace('.', '-')}"
            if bucket not in config.k.list_buckets():
                config.k.create_bucket(bucket)
            config.k.upload_to_bucket(bucket, f"{clusterdir}/nodepool_{cluster}.ign", public=True)
            bucket_url = config.k.public_bucketfile_url(bucket, "nodepool_{cluster}.ign")
            new_ignition = {'ignition': {'config': {'merge': [{'source': bucket_url}]}, 'version': '3.2.0'}}
            with open(f"{clusterdir}/nodepool_{cluster}.ign", 'w') as f:
                f.write(json.dumps(new_ignition))
        elif baremetal_iso or baremetal_hosts:
            iso_url = handle_baremetal_iso(config, plandir, cluster, data, baremetal_hosts)
            result = start_baremetal_hosts_with_iso(baremetal_hosts, iso_url, overrides=overrides, debug=config.debug)
            if result['result'] != 'success':
                return result
            data['workers'] = data.get('workers', 2) - len(baremetal_hosts)
    pprint("Waiting for kubeconfig to be available")
    call(f"until oc -n {namespace} get secret {cluster}-admin-kubeconfig >/dev/null 2>&1 ; do sleep 1 ; done",
         shell=True)
    kubeconfigpath = f'{clusterdir}/auth/kubeconfig'
    kubeconfig = os.popen(f"oc extract -n {namespace} secret/{cluster}-admin-kubeconfig --to=-").read()
    with open(kubeconfigpath, 'w') as f:
        f.write(kubeconfig)
    pprint("Waiting for kubeadmin-password to be available")
    call(f"until oc -n {namespace} get secret {cluster}-kubeadmin-password >/dev/null 2>&1 ; do sleep 1 ; done",
         shell=True)
    kubeadminpath = f'{clusterdir}/auth/kubeadmin-password'
    kubeadmin = os.popen(f"oc extract -n {namespace} secret/{cluster}-kubeadmin-password --to=-").read()
    with open(kubeadminpath, 'w') as f:
        f.write(kubeadmin)
    if network_type == 'Calico':
        calico_version = data['calico_version']
        with TemporaryDirectory() as tmpdir:
            calico_data = {'tmpdir': tmpdir, 'namespace': namespace, 'cluster': cluster, 'clusterdir': clusterdir,
                           'calico_version': calico_version}
            calico_script = config.process_inputfile('xxx', f'{plandir}/calico.sh.j2', overrides=calico_data)
            with open(f"{tmpdir}/calico.sh", 'w') as f:
                f.write(calico_script)
            call(f'bash {tmpdir}/calico.sh', shell=True)
    elif network_type == 'Cilium':
        cilium_version = data['cilium_version']
        cluster_network_ipv4 = data['cluster_network_ipv4']
        cilium_data = {'clusterdir': clusterdir, 'cilium_version': cilium_version, 'cidr': cluster_network_ipv4}
        cilium_script = config.process_inputfile('xxx', f'{plandir}/cilium.sh.j2', overrides=cilium_data)
        with open(f"{clusterdir}/cilium.sh", 'w') as f:
            f.write(cilium_script)
        call(f'bash {clusterdir}/cilium.sh', shell=True)
    if platform is None and data['workers'] > 0:
        pprint("Deploying workers")
        worker_threaded = data.get('threaded', False) or data.get('workers_threaded', False)
        result = config.plan(plan, inputfile=f'{plandir}/kcli_plan.yml', overrides=data, threaded=worker_threaded)
        if result['result'] != 'success':
            return result
    async_tempdir.cleanup()
    if ignore_hosts:
        warning("Not updating /etc/hosts as per your request")
    elif ingress_ip is not None:
        update_hypershift_etc_hosts(cluster, domain, ingress_ip)
    if provider in cloud_providers:
        result = config.plan(plan, inputfile=f'{plandir}/cloud_lb_apps.yml', overrides=data)
        if result['result'] != 'success':
            return result
    if async_install or which('openshift-install') is None:
        success(f"Kubernetes cluster {cluster} deployed!!!")
        info2(f"export KUBECONFIG=$HOME/.kcli/clusters/{cluster}/auth/kubeconfig")
        info2("export PATH=$PWD:$PATH")
    else:
        installcommand = f'openshift-install --dir={clusterdir} --log-level={log_level} wait-for install-complete'
        installcommand = ' || '.join([installcommand for x in range(retries)])
        pprint("Launching install-complete step. It will be retried extra times to handle timeouts")
        run = call(installcommand, shell=True)
        if run != 0:
            msg = "Leaving environment for debugging purposes. "
            msg += f"Delete it with kcli delete kube --yes {cluster}"
            return {'result': 'failure', 'reason': msg}
    if provider in cloud_providers:
        bucket = f"{cluster}-{domain.replace('.', '-')}"
        config.k.delete_bucket(bucket)
    os.environ['KUBECONFIG'] = f"{clusterdir}/auth/kubeconfig"
    apps = overrides.get('apps', [])
    overrides['hypershift'] = True
    overrides['cluster'] = cluster
    if ingress_metallb:
        if 'metallb-operator' not in apps:
            apps.append('metallb-operator')
        if 'metallb_ranges' not in overrides:
            overrides['metallb_ranges'] = [ingress_ip]
        ingress_file = config.process_inputfile('xxx', f'{plandir}/assisted_ingress_service.yml.j2',
                                                overrides={'ingress_ip': ingress_ip})
        with open(f"{clusterdir}/assisted_ingress_service.yml", 'w') as f:
            f.write(ingress_file)
        call(f'oc create -f {clusterdir}/assisted_ingress_service.yml', shell=True)
    process_apps(config, clusterdir, apps, overrides)
    if autoscale:
        config.import_in_kube(network=network, secure=True)
        with NamedTemporaryFile(mode='w+t') as temp:
            commondir = os.path.dirname(pprint.__code__.co_filename)
            autoscale_overrides = {'cluster': cluster, 'kubetype': 'k3s', 'workers': workers, 'replicas': 1,
                                   'sa': 'default'}
            autoscale_data = config.process_inputfile(cluster, f"{commondir}/autoscale.yaml.j2",
                                                      overrides=autoscale_overrides)
            temp.write(autoscale_data)
            temp.seek(0)
            scc_cmd = "oc adm policy add-scc-to-user anyuid -z default -n kcli-infra"
            call(scc_cmd, shell=True)
            autoscale_cmd = f"oc create -f {temp.name}"
            call(autoscale_cmd, shell=True)
    if provider in cloud_providers and data.get('cloud_storage', True):
        pprint("Deploying cloud storage class")
        deploy_cloud_storage(config, cluster, apply=False)
    return {'result': 'success'}
